from orionis.container.facades.facade import Facade

class PerformanceCounter(Facade):

    @classmethod
    def getFacadeAccessor(cls):
        """
        Returns the registered binding key for the performance counter service in the service container.

        This method provides the unique string identifier used by the service container to resolve
        and retrieve the implementation of the performance counter service. It acts as the connection
        point between the facade and the underlying service registration, ensuring that the correct
        service is accessed when requested through the facade.

        Returns
        -------
        str
            The binding key 'x-orionis.support.performance.contracts.counter.IPerformanceCounter'
            which identifies the performance counter service implementation in the service container.
        """

        # Return the unique binding key for the performance counter service
        return "x-orionis.support.performance.contracts.counter.IPerformanceCounter"


INSPIRATIONAL_QUOTES = [
    {"quote": "Bitterness is like cancer. It eats upon the host.", "author": "Maya Angelou"},
    {"quote": "The true test of character is not how much we know how to do, but how we behave when we don’t know what to do.", "author": "John Holt"},
    {"quote": "It's better to be at the bottom of a ladder you want to climb than halfway up one you don't.", "author": "Stephen Kellogg"},
    {"quote": "The biggest temptation is to settle for too little.", "author": "Thomas Merton"},
    {"quote": "Everybody wants to be famous, but nobody wants to do the work. I live by that. You grind hard so you can play hard. At the end of the day, you put all the work in, and eventually it’ll pay off. It could be in a year, it could be in 30 years. Eventually, your hard work will pay off.", "author": "Kevin Hart"},
    {"quote": "Every action you take is a vote for the type of person you wish to become.", "author": "James Clear"},
    {"quote": "You pray for the hungry. Then you feed them. That’s how prayer works.", "author": "Pope Francis"},
    {"quote": "I would like to show the world today as an ant sees it, and tomorrow as the moon sees it.", "author": "Hannah Höch"},
    {"quote": "Write this down, young man. Life has been extremely, I say extremely, kind.", "author": "Margaret Gorman"},
    {"quote": "What you are, you are by accident of birth; what I am, I am by myself.", "author": "Ludwig van Beethoven"},
    {"quote": "Never allow a person to tell you no who doesn’t have the power to say yes.", "author": "Eleanor Roosevelt"},
    {"quote": "When I hear somebody sigh, 'Life is hard,' I am always tempted to ask, 'Compared to what?'", "author": "Sydney J. Harris"},
    {"quote": "If you do every job like you’re going to do it for the rest of your life, that’s when you get noticed.", "author": "Mary Barra"},
    {"quote": "Anyone who has spent any time in space will love it for the rest of their lives. I achieved my childhood dream of the sky.", "author": "Valentina Tereshkova"},
    {"quote": "Your messaging should convey, 'Here’s why you would be so lucky if you got me,' instead of 'I really wanted it,' which we all tend to do.", "author": "Barbara Corcoran"},
    {"quote": "All medicine is like that. It came from someone who dared.", "author": "Dr. William DeVries"},
    {"quote": "The American city should be a collection of communities where every member has a right to belong. It should be a place where every man feels safe on his streets and in the house of his friends.", "author": "Lyndon B. Johnson"},
    {"quote": "I knew all the rules but the rules did not know me.", "author": "Eddie Vedder"},
    {"quote": "It was the first evidence to me that we’d changed the world.", "author": "John Shepherd-Barron"},
    {"quote": "The things that scare you only make you better and stronger.", "author": "Octavia Spencer"},
    {"quote": "The more I want to get something done, the less I call it work.", "author": "Richard Bach"},
    {"quote": "It's a hard, lonely feeling, to be completely yourself in front of strangers.", "author": "Michael Che"},
    {"quote": "The first step is clearly defining what it is you’re after, because without knowing that, you’ll never get it.", "author": "Halle Berry"},
    {"quote": "Take the opportunity to learn from your mistakes: find the cause of your problem and eliminate it. Don’t try to be perfect; just be an excellent example of being human.", "author": "Tony Robbins"},
    {"quote": "I'm not the next Usain Bolt or Michael Phelps … I'm the first Simone Biles.", "author": "Simone Biles"},
    {"quote": "My mother thinks I am the best. And I was raised to always believe what my mother tells me.", "author": "Diego Maradona"},
    {"quote": "You would be amazed what the ordinary guy knows.", "author": "Matt Drudge"},
    {"quote": "You are never too good for anything. Always be willing to do everything.", "author": "Thomas Parkinson"},
    {"quote": "Do as much homework as you can. Learn everybody’s job and don’t just settle.", "author": "Michael B. Jordan"},
    {"quote": "Those who don’t believe in magic will never find it.", "author": "Roald Dahl"},
    {"quote": "There is no such thing as accident; it is fate misnamed.", "author": "Napoleon Bonaparte"},
    {"quote": "The function of leadership is to produce more leaders, not more followers.", "author": "Ralph Nader"},
    {"quote": "You can overcome anything, if and only if you love something enough.", "author": "Lionel Messi"},
    {"quote": "God works a miracle!", "author": "William Bradford"},
    {"quote": "The man who moves a mountain begins by carrying away small stones.", "author": "Confucius"},
    {"quote": "To make others less happy is a crime. To make ourselves unhappy is where all crime starts.", "author": "Roger Ebert"},
    {"quote": "Opportunities multiply as they are seized.", "author": "Sun Tzu"},
    {"quote": "People who think they know everything are a great annoyance to those of us who do.", "author": "Isaac Asimov"},
    {"quote": "Give me a place to stand, and a lever long enough, and I will move the world.", "author": "Archimedes"},
    {"quote": "People will forget what you said, people will forget what you did, but people will never forget how you made them feel.", "author": "Maya Angelou"},
    {"quote": "I am not afraid of an army of lions led by a sheep; I am afraid of an army of sheep led by a lion.", "author": "Alexander the Great"},
    {"quote": "We're born alone. We do need each other. It's lonely to really effectively live your life, and anyone you can get help from or give help to, that's part of your obligation.", "author": "Bill Murray"},
    {"quote": "Failure is a badge of honor. It means you risked failure. And if you don’t risk failure, you’re never going to do anything that’s different from what you’ve already done or what somebody else has done.", "author": "Charlie Kaufman"},
    {"quote": "This is not just a job. This is a passion. Connecting people, connecting the world, making a difference. How many jobs can say I made the world better today? We do that every single day.", "author": "Ed Bastian"},
    {"quote": "A delayed game is eventually good, but a rushed game is forever bad.", "author": "Shigeru Miyamoto"},
    {"quote": "It's a gift to exist, and with existence comes suffering. There's no escaping that.", "author": "Stephen Colbert"},
    {"quote": "I believe ambition is not a dirty word. It's just believing in yourself and your abilities. Imagine this: What would happen if we were all brave enough to be a little bit more ambitious? I think the world would change.", "author": "Reese Witherspoon"},
    {"quote": "Certain things in life are more important than the usual crap that everyone strives for.", "author": "James Taylor"},
    {"quote": "Logic will get you from A to B. Imagination will take you everywhere.", "author": "Albert Einstein"},
    {"quote": "If you feel safe in the area you’re working in, you’re not working in the right area. Always go a little further into the water than you feel you’re capable of being in. Go a little bit out of your depth. And when you don’t feel that your feet are quite touching the bottom, you’re just about in the right place to do something exciting.", "author": "David Bowie"},
    {"quote": "If you can't fail, you can never get better.", "author": "Benedict Cumberbatch"},
    {"quote": "It took me a long time not to judge myself through someone else’s eyes.", "author": "Sally Field"},
    {"quote": "The best protection any woman can have … is courage.", "author": "Elizabeth Cady Stanton"},
    {"quote": "I have captured the light and arrested its flight. The sun itself shall draw my pictures.", "author": "Louis Daguerre"},
    {"quote": "It's beautiful when you watch something good happen to somebody when it's well deserved.", "author": "Jennifer Lawrence"},
    {"quote": "You can't teach anybody anything, only make them realize the answers are already inside them.", "author": "Galileo Galilei"},
    {"quote": "Talk sense to a fool and he calls you foolish.", "author": "Euripides"},
    {"quote": "If you love what you do and are willing to do what it takes, it’s within your reach.", "author": "Steve Wozniak"},
    {"quote": "Troubles are often the tools by which God fashions us for better things.", "author": "Henry Ward Beecher"},
    {"quote": "To love means loving the unlovable. To forgive means pardoning the unpardonable. Faith means believing the unbelievable. Hope means hoping when everything seems hopeless.", "author": "Gilbert K. Chesterton"},
    {"quote": "The world promises you comfort, but you were not made for comfort. You were made for greatness.", "author": "Pope Benedict XVI"},
    {"quote": "Memory is original events combined with every other time you’ve remembered that event. This is why therapy’s helpful.", "author": "Becky Kennedy"},
    {"quote": "If opportunity doesn’t knock, build a door.", "author": "Milton Berle"},
    {"quote": "It's easy to fool the eye but it's hard to fool the heart.", "author": "Al Pacino"},
    {"quote": "It is not the strongest of the species that survive, nor the most intelligent, but the one most responsive to change.", "author": "Charles Darwin"},
    {"quote": "Success is getting what you want. Happiness is wanting what you get.", "author": "Dale Carnegie"},
    {"quote": "You are on the eve of a complete victory. You can’t go wrong. The world is behind you.", "author": "Josephine Baker"},
    {"quote": "Holding on to anger is like grasping a hot coal with the intent of throwing it at someone else; you are the one who gets burned.", "author": "Buddha"},
    {"quote": "Criticism is something we can avoid easily by saying nothing, doing nothing, and being nothing.", "author": "Aristotle"},
    {"quote": "Don’t let what you don’t know scare you, because it can become your greatest asset. And if you do things without knowing how they have always been done, you’re guaranteed to do them differently.", "author": "Sara Blakely"},
    {"quote": "An innovation will get traction only if it helps people get something that they’re already doing in their lives done better.", "author": "Clayton Christensen"},
    {"quote": "Nursing encompasses an art, a humanistic orientation, a feeling for the value of the individual, and an intuitive sense of ethics, and of the appropriateness of action taken.", "author": "Myrtle Aydelotte"},
    {"quote": "It seems to me that to be peaceful and content, you don’t have to know precisely what you want. It may even be better if you don’t. But I think you do need to be very clear what you DON’T want, and steadfast in the not-doing of those things.", "author": "Sandra Boynton"},
    {"quote": "It was the ultimate dream. But the idea of a young black American like me owning a Cadillac? It seemed too far-fetched.", "author": "Dick Gidron"},
    {"quote": "You are a Universe of Universes and your Soul a source of Songs.", "author": "Rubén Darío"},
    {"quote": "Don’t live the same year 75 times and call it a life.", "author": "Robin Sharma"},
    {"quote": "If there is no struggle, there is no progress.", "author": "Frederick Douglass"},
    {"quote": "Think about me every now and then, old friend.", "author": "John Lennon"},
    {"quote": "We must never be afraid to go too far, for success lies just beyond.", "author": "Marcel Proust"},
    {"quote": "There is no list of rules. There is one rule. The rule is: There are no rules.", "author": "Shonda Rhimes"},
    {"quote": "I would rather have one article a day of this sort; and these 10 or 20 lines might readily represent a whole day’s hard work in the way of concentrated, intense thinking and revision, polish of style, weighing of words.", "author": "Joseph Pulitzer"},
    {"quote": "Winners are the ones who really listen to the truth of their hearts.", "author": "Sylvester Stallone"},
    {"quote": "With everything that has happened to you, you can either feel sorry for yourself or treat what has happened as a gift. Everything is either an opportunity to grow or an obstacle to keep you from growing. You get to choose.", "author": "Wayne Dyer"},
    {"quote": "If you’re too rigid in your thinking you may miss some wonderful opportunities for storytelling.", "author": "Vince Gilligan"},
    {"quote": "Education is not the filling of a pail, but the lighting of a fire.", "author": "William Butler Yeats"},
    {"quote": "Before you can make a dream come true, you must first have one.", "author": "Ronald McNair"},
    {"quote": "Business is built on relationships, relationships are built on trust, and the foundation of trust is honesty.", "author": "Stephen Covey"},
    {"quote": "Far better is it to dare mighty things, to win glorious triumphs, even though checkered by failure, than to rank with those poor spirits who neither enjoy much nor suffer much, because they live in a gray twilight that knows not victory nor defeat.", "author": "Theodore Roosevelt"},
    {"quote": "What we do is more important than what we say or what we say we believe.", "author": "bell hooks"},
    {"quote": "Sometimes all you need is 20 seconds of insane courage.", "author": "Benjamin Mee"},
    {"quote": "Be passionate and bold. Always keep learning. You stop doing useful things if you don’t learn.", "author": "Satya Nadella"},
    {"quote": "The chief cause of failure is substituting what you want most for what you want now.", "author": "Zig Ziglar"},
    {"quote": "If you know exactly what you want to be, you need to spend as much time as possible with people who are actually that already.", "author": "Gary Vaynerchuk"},
    {"quote": "You never know what worse luck your bad luck has saved you from.", "author": "Cormac McCarthy"},
    {"quote": "Leadership is the art of getting someone else to do something you want done because he wants to do it.", "author": "Dwight D. Eisenhower"},
    {"quote": "If it wasn’t hard, everyone would do it. It’s the hard that makes it great.", "author": "Tom Hanks"},
    {"quote": "Sometimes your mistakes are your biggest virtues. You learn so much from the mistake. Those things that you think are the worst thing that’s happening to you can somehow turn around and be the greatest opportunity.", "author": "Nicole Kidman"},
    {"quote": "Get into what your kids are into. How miserable would it be to always be anti-popular stuff like 'Ugh, you seriously like Taylor Swift?'", "author": "Joel Willis"},
    {"quote": "Life is not easy for any of us. But what of that? We must have perseverance and, above all, confidence in ourselves. We must believe we are gifted for something and that this thing must be attained.", "author": "Marie Curie"},
    {"quote": "We hold our heads high, despite the price we have paid, because freedom is priceless.", "author": "Lech Wałęsa"},
    {"quote": "It's always better to shock people and change people's expectations than to give them exactly what they think you can do.", "author": "Jonah Hill"},
    {"quote": "I think happiness is overrated. Satisfied, at peace—those would be more realistic goals.", "author": "Brad Pitt"},
    {"quote": "It is not very often that an opportunity comes knocking. But when it does, you’d better be bathed and dressed and ready to answer its call.", "author": "Jyoti Arora"},
    {"quote": "To understand the heart and mind of a person, look not at what he has already achieved, but at what he aspires to do.", "author": "Kahlil Gibran"},
    {"quote": "Diligence is the mother of good luck.", "author": "Benjamin Franklin"},
    {"quote": "The rung of a ladder was never meant to rest upon, but only to hold a man’s foot long enough to enable him to put the other somewhat higher.", "author": "Thomas Huxley"},
    {"quote": "Thousands of candles can be lighted from a single candle. Happiness never decreases by being shared.", "author": "Buddha"},
    {"quote": "It's hard work to make a four-minute program look effortless and elegant.", "author": "Katarina Witt"},
    {"quote": "I would rather five people knew my work and thought it was good work than five million knew me and were indifferent.", "author": "Colin Firth"},
    {"quote": "Someday is not a day of the week.", "author": "Janet Dailey"},
    {"quote": "If everything seems under control, you’re not going fast enough.", "author": "Mario Andretti"},
    {"quote": "An important attribute of success is to be yourself. Never hide what makes you, you.", "author": "Indra Nooyi"},
    {"quote": "What is harder than rock, or softer than water? Yet soft water hollows out hard rock. Persevere.", "author": "Ovid"},
    {"quote": "My pain may be the reason for somebody’s laugh. But my laugh must never be the reason for somebody’s pain.", "author": "Charlie Chaplin"},
    {"quote": "Don’t worry about failure; you only have to be right once.", "author": "Drew Houston"},
    {"quote": "A lot of times, people look at the negative side of what they feel they can’t do. I always look on the positive side of what I can do.", "author": "Chuck Norris"},
    {"quote": "I tend to think you’re fearless when you recognize why you should be scared of things, but do them anyway.", "author": "Christian Bale"},
    {"quote": "Success is built on the foundation of courage, hard work and individual responsibility. Despite what some would have us believe, success is not built on resentment and fears.", "author": "Susana Martinez"},
    {"quote": "I practice my saxophone three hours a day. I’m not saying I’m particularly special, but if you do something three hours a day for forty years, you get pretty good at it.", "author": "Kenny G"},
    {"quote": "Smooth seas do not make skillful sailors.", "author": "African Proverb"},
    {"quote": "I know the price of success: dedication, hard work and an unremitting devotion to the things you want to see happen.", "author": "Frank Lloyd Wright"},
    {"quote": "I think one day you’ll find that you’re the hero you’ve been looking for.", "author": "Jimmy Stewart"},
    {"quote": "If all I do in my life is soothe someone’s spirit with a song, then let me do that and I’m happy.", "author": "Gladys Knight"},
    {"quote": "Life becomes easier when you learn to accept the apology you never got.", "author": "Robert Brault"},
    {"quote": "There are times in all of our lives when a reliance on gut or intuition just seems more appropriate, when a particular course of action just feels right. And, interestingly, I’ve discovered it’s in facing life’s most important decisions that intuition seems the most indispensable.", "author": "Tim Cook"},
    {"quote": "All labor that uplifts humanity has dignity and importance.", "author": "Martin Luther King Jr."},
    {"quote": "You may think using Google’s great, but I still think it’s terrible.", "author": "Larry Page"},
    {"quote": "It’s tough to get out of bed to do roadwork at 5 a.m. when you’ve been sleeping in silk pajamas.", "author": "Marvin Hagler"},
    {"quote": "There is no security on the Earth, there is only opportunity.", "author": "Douglas MacArthur"},
    {"quote": "Everyone’s like, 'overnight sensation.' It’s not overnight. It’s years of hard work.", "author": "Margot Robbie"},
    {"quote": "The worth of a life is not determined by a single failure or a solitary success.", "author": "Kevin Kline"},
    {"quote": "Only the weak are cruel. Gentleness can only be expected from the strong.", "author": "Leo Buscaglia"},
    {"quote": "For small creatures such as we, the vastness is bearable only through love.", "author": "Carl Sagan"},
    {"quote": "Build something 100 people love, not something 1 million people kind of like.", "author": "Brian Chesky"},
    {"quote": "Once you replace negative thoughts with positive ones, you’ll start having positive results.", "author": "Willie Nelson"},
    {"quote": "Don’t write what you think people want to read. Find your voice and write about what’s in your heart.", "author": "Quentin Tarantino"},
    {"quote": "Life is what we make it, always has been, always will be.", "author": "Grandma Moses"},
    {"quote": "There is overwhelming evidence that the higher the level of self-esteem, the more likely one will be to treat others with respect, kindness, and generosity.", "author": "Nathaniel Branden"},
    {"quote": "It’s a good place when all you have is hope and not expectations.", "author": "Danny Boyle"},
    {"quote": "Good things come to obsessive compulsives who fixate.", "author": "Kieran Culkin"},
    {"quote": "It is easier to build strong children than to repair broken men.", "author": "Frederick Douglass"},
    {"quote": "You can’t be afraid of what people are going to say, because you’re never going to make everyone happy.", "author": "Selena Gomez"},
    {"quote": "Waking up in truth is so much better than living in a lie.", "author": "Idris Elba"},
    {"quote": "Instead of looking at the past, I put myself ahead 20 years and try to look at what I need to do now in order to get there then.", "author": "Diana Ross"},
    {"quote": "People seldom do what they believe in. They do what is convenient, and then repent.", "author": "Bob Dylan"},
    {"quote": "Anything in life worth having is worth working for.", "author": "Andrew Carnegie"},
    {"quote": "Life is short, and the world is wide.", "author": "Simon Raven"},
    {"quote": "I felt that one had better die fighting against injustice than to die like a dog or rat in a trap. I had already determined to sell my life as dearly as possible if attacked. I felt if I could take one lyncher with me, this would even up the score a little bit.", "author": "Ida B. Wells"},
    {"quote": "There are two ways of spreading light: to be the candle or the mirror that reflects it.", "author": "Edith Wharton"},
    {"quote": "How old would you be if you didn’t know how old you are?", "author": "Satchel Paige"},
    {"quote": "When you find an idea that you just can’t stop thinking about, that’s probably a good one to pursue.", "author": "Josh James"},
    {"quote": "We each come by the gifts we have to offer by an infinite series of influences and lucky breaks we can never fully understand.", "author": "MacKenzie Scott"},
    {"quote": "Everyone thinks of changing the world, but no one thinks of changing himself.", "author": "Leo Tolstoy"},
    {"quote": "Economics is like the Dutch language. I’m told it makes sense, but I have my doubts.", "author": "John Oliver"},
    {"quote": "Ultimately, you just have one life. You never know unless you try. And you never get anywhere unless you ask.", "author": "Kate Winslet"},
    {"quote": "We are all ordinary. We are all boring. We are all spectacular. We are all shy. We are all bold. We are all heroes. We are all helpless. It just depends on the day.", "author": "Brad Meltzer"},
    {"quote": "Think of how stupid the average person is, and realize half of them are stupider than that.", "author": "George Carlin"},
    {"quote": "Some of the world’s greatest feats were accomplished by people not smart enough to know they were impossible.", "author": "Doug Larson"},
    {"quote": "Take wrong turns. Talk to strangers. Open unmarked doors. And if you see a group of people in a field, go find out what they are doing. Do things without always knowing how they’ll turn out.", "author": "Randall Munroe"},
    {"quote": "I’m intimidated by the fear of being average.", "author": "Taylor Swift"},
    {"quote": "Always be a first-rate version of yourself instead of a second-rate version of somebody else.", "author": "Judy Garland"},
    {"quote": "The history of the world is the history of a few people who had faith in themselves.", "author": "Swami Vivekananda"},
    {"quote": "The one thing you’re most reluctant to tell. That’s where the comedy is.", "author": "Mike Birbiglia"},
    {"quote": "Doubt is a killer. You just have to know who you are and what you stand for.", "author": "Jennifer Lopez"},
    {"quote": "Basically, the first half of life is writing the text, and the second half is writing the commentary on that text.", "author": "Richard Rohr"},
    {"quote": "Great artists steal is about finding inspiration in the work of others, then using it as a starting point for original creative output.", "author": "Pablo Picasso"},
    {"quote": "Don’t count the days, make the days count.", "author": "Muhammad Ali"},
    {"quote": "You must not only aim right, but draw the bow with all your might.", "author": "Henry David Thoreau"},
    {"quote": "Your days are numbered. Use them to throw open the windows of your soul to the sun. If you do not, the sun will soon set, and you with it.", "author": "Marcus Aurelius"},
    {"quote": "I have already lost touch with a couple of people I used to be.", "author": "Joan Didion"},
    {"quote": "Fear of looking stupid is the No. 1 killer of dreams. The worst part? The people who make you feel stupid are usually the ones least qualified to judge someone else’s life.", "author": "Anthony Moore"},
    {"quote": "I’ve grown most not from victories but from setbacks. If winning is God’s reward, then losing is how he teaches us.", "author": "Serena Williams"},
    {"quote": "Life is 10 percent what happens to me and 90 percent of how I react to it.", "author": "Charles R. Swindoll"},
    {"quote": "A successful man is one who can lay a firm foundation with the bricks others have thrown at him.", "author": "David Brinkley"},
    {"quote": "I don’t know what the future may hold, but I know who holds the future.", "author": "Ralph Abernathy"},
    {"quote": "People always fall in love with the most perfect aspects of each other’s personalities. Who wouldn’t? Anybody can love the most wonderful parts of another person. But that’s not the clever trick. The really clever trick is this: Can you accept the flaws?", "author": "Elizabeth Gilbert"},
    {"quote": "Being an optimist after you’ve got the very thing you want doesn’t count.", "author": "Ken Hubbard"},
    {"quote": "I can say the willingness to get dirty has always defined us as a nation, and it’s a hallmark of hard work and a hallmark of fun, and dirt is not the enemy.", "author": "Mike Rowe"},
    {"quote": "I’ll tell you what bravery really is. Bravery is just determination to do a job that you know has to be done.", "author": "Audie Murphy"},
    {"quote": "Always look for the fool in the deal. If you don’t find one, it’s you.", "author": "Mark Cuban"},
    {"quote": "If everyone is thinking alike, then somebody isn’t thinking.", "author": "George S. Patton"},
    {"quote": "No matter what people tell you, words and ideas can change the world.", "author": "Robin Williams"},
    {"quote": "Philosophy is like trying to open a safe with a combination lock: each little adjustment of the dials seems to achieve nothing, only when everything is in place does the door open.", "author": "Ludwig Wittgenstein"},
    {"quote": "I am not a product of my circumstances. I am a product of my decisions.", "author": "Stephen Covey"},
    {"quote": "Victory has a hundred fathers and defeat is an orphan.", "author": "John F. Kennedy"},
    {"quote": "If you hear the dogs, keep going. If you see the torches in the woods, keep going. If there’s shouting after you, keep going. Don’t ever stop. Keep going. If you want a taste of freedom, keep going.", "author": "Harriet Tubman"},
    {"quote": "Art, in itself, is an attempt to bring order out of chaos.", "author": "Stephen Sondheim"},
    {"quote": "Reality is created by the mind; we can change our reality by changing our mind.", "author": "Plato"},
    {"quote": "Grant me courage to serve others; for in service there is true life.", "author": "Cesar Chavez"},
    {"quote": "Anyone with two tunics should share with him who has none.", "author": "John the Baptist"},
    {"quote": "A pessimist sees the difficulty in every opportunity; an optimist sees the opportunity in every difficulty.", "author": "Winston Churchill"},
    {"quote": "What I’ve learned from running is that the time to push hard is when you’re hurting like crazy and you want to give up. Success is often just around the corner.", "author": "James Dyson"},
    {"quote": "Do you really want to look back on your life and see how wonderful it could have been had you not been afraid to live it?", "author": "Caroline Myss"},
    {"quote": "The secret of a happy marriage is finding the right person. You know they’re right if you love to be with them all the time.", "author": "Julia Child"},
    {"quote": "Flaming enthusiasm, backed by horse-sense and persistence, is the quality that most frequently makes for success.", "author": "Dale Carnegie"},
    {"quote": "God loves us beyond comprehension, and we cannot diminish God’s love for us.", "author": "Peter Kreeft"},
    {"quote": "What gets celebrated gets replicated.", "author": "Bradley Cooper"},
    {"quote": "The best way to predict the future is to invent it.", "author": "Alan Kay"},
    {"quote": "Everyone you meet in life is fighting a battle you know nothing about.", "author": "Wendy Mass"},
    {"quote": "If you look around the room, and you’re the smartest person in the room, you’re in the wrong room.", "author": "Lorne Michaels"},
    {"quote": "Being polite and grateful will make people more inclined to help you. And if people are willing to help you, you may accidentally get something you want.", "author": "Jason Sudeikis"},
    {"quote": "If people knew how hard I worked to achieve my mastery, it wouldn’t seem so wonderful after all.", "author": "Michelangelo"},
    {"quote": "If you want to change who you are, you have to change what you do.", "author": "Jude Law"},
    {"quote": "The best time to plant a tree was 20 years ago. The second best time is now.", "author": "Chinese Proverb"},
    {"quote": "Bridge the age gap. For younger people, find a way to spend time with older people. For the more silver-haired among us, I’m telling you: Find a few things the young kids are into — music, tech, sports — and check them out.", "author": "Mike Erwin"},
    {"quote": "I’m not afraid of storms, for I’m learning how to sail my ship.", "author": "Louisa May Alcott"},
    {"quote": "Today a reader, tomorrow a leader.", "author": "Margaret Fuller"},
    {"quote": "Freedom lies in being bold.", "author": "Robert Frost"},
    {"quote": "I will take the subway and look at certain women and think 'God, that woman’s story will never be told. How come that lady doesn’t get a movie about her?'", "author": "Natasha Lyonne"},
    {"quote": "If you focus on results you’ll never change. If you focus on change, you’ll get results.", "author": "Jack Dixon"},
    {"quote": "Empty your mind, be formless. Shapeless, like water. If you put water into a cup, it becomes the cup. You put water into a bottle and it becomes the bottle. You put it in a teapot it becomes the teapot. Now, water can flow or it can crash. Be water, my friend.", "author": "Bruce Lee"},
    {"quote": "Complexity kills. It sucks the life out of developers, it makes products difficult to plan, build and test, it introduces security challenges, and it causes end-user and administrator frustration.", "author": "Ray Ozzie"},
    {"quote": "When one door closes another door opens; but we so often look so long and so regretfully upon the closed door, that we do not see the ones which open for us.", "author": "Alexander Graham Bell"},
    {"quote": "The greatest lesson that I learned in all of this is that you have to start. Start now, start here, and start small. Keep it simple.", "author": "Jack Dorsey"},
    {"quote": "The crisis of today is the joke of tomorrow.", "author": "H.G. Wells"},
    {"quote": "Either America will destroy ignorance or ignorance will destroy the United States.", "author": "W. E. B. Du Bois"},
    {"quote": "It would be wonderful if I can inspire others, who are struggling to realize their dreams, to say: If this country kid could do it, let me keep slogging away.", "author": "Douglas Engelbart"},
    {"quote": "Everything you’ve ever wanted is on the other side of fear.", "author": "George Addair"},
    {"quote": "Sometimes you can only find heaven by slowly backing away from hell.", "author": "Carrie Fisher"},
    {"quote": "I’ve got to give props to the Jesuit priests.", "author": "Neil deGrasse Tyson"},
    {"quote": "If God took the trouble to tell us eight hundred times to be glad and rejoice, He must want us to do it!", "author": "Eleanor H. Porter"},
    {"quote": "That’s the best part of the whole thing, is the people that we’ve met along the way. We’ve met some wonderful, wonderful people and we wouldn’t have otherwise.", "author": "Pearl Dion"},
    {"quote": "Conditions are never perfect. 'Someday' is a disease that will take your dreams to the grave with you. If it’s important to you and you want to do it 'eventually,' just do it and correct course along the way.", "author": "Tim Ferriss"},
    {"quote": "Whatever you do, don’t wake up at 65 years old and think about what you should have done with your life.", "author": "George Clooney"},
    {"quote": "My soul is that of a drummer. I didn’t do it to become rich and famous. I did it because it was the love of my life.", "author": "Ringo Starr"},
    {"quote": "Just keep going. Everybody gets better if they keep at it.", "author": "Ted Williams"},
    {"quote": "There can be no greater gift than that of giving one’s time and energy to help others without expecting anything in return.", "author": "Nelson Mandela"},
    {"quote": "Good luck has its storms.", "author": "George Lucas"},
    {"quote": "Faith moves mountains, but you have to keep pushing while you’re praying.", "author": "Mason Cooley"},
    {"quote": "No failure means no risk, which means nothing new.", "author": "Richard Branson"},
    {"quote": "I’m a big believer in the power of inexperience. It was the greatest asset I had when I started TFA. If I had known at the outset how hard it was going to be, I might never have started.", "author": "Wendy Kopp"},
    {"quote": "Who sows virtue reaps honor.", "author": "Leonardo da Vinci"},
    {"quote": "So often, people are working hard at the wrong thing. Working on the right thing is probably more important than working hard.", "author": "Caterina Fake"},
    {"quote": "I think of doing a series as very hard work. But then I’ve talked to coal miners, and that’s really hard work.", "author": "William Shatner"},
    {"quote": "Life is a daring adventure or it is nothing at all.", "author": "Helen Keller"},
    {"quote": "Whenever you read a good book, somewhere in the world a door opens to allow in more light.", "author": "Vera Nazarian"},
    {"quote": "To see we must forget the name of the thing we are looking at.", "author": "Claude Monet"},
    {"quote": "I didn’t like the idea of being foolish, but I learned pretty soon that it was essential to fail and be foolish.", "author": "Daniel Day-Lewis"},
    {"quote": "I hate that word: lucky. It cheapens a lot of hard work.", "author": "Peter Dinklage"},
    {"quote": "Even when I was close to defeat I rose to my feet.", "author": "Dr. Dre"},
    {"quote": "Happiness is an inside job.", "author": "William Arthur Ward"},
    {"quote": "I was literally just doing my job.", "author": "Stanislav Petrov"},
    {"quote": "All good ideas start out as bad ideas, that’s why it takes so long.", "author": "Steven Spielberg"},
    {"quote": "The world has never had a good definition of the word liberty, and the American people, just now, are much in want of one.", "author": "Abraham Lincoln"},
    {"quote": "I wanted a perfect ending. Now I’ve learned, the hard way, that some poems don’t rhyme, and some stories don’t have a clear beginning, middle and end.", "author": "Gilda Radner"},
    {"quote": "Giving up smoking is the easiest thing in the world. I know because I’ve done it thousands of times.", "author": "Mark Twain"},
    {"quote": "Only eyes washed by tears can see clearly.", "author": "Louis Mann"},
    {"quote": "You have to learn the rules to be able to know how to break them.", "author": "Keira Knightley"},
    {"quote": "The meaning of life is to find your gift. The purpose of life is to give it away.", "author": "Pablo Picasso"},
    {"quote": "You are never too old to set another goal or to dream a new dream.", "author": "Les Brown"},
    {"quote": "The essence of strategy is choosing what not to do.", "author": "Michael Porter"},
    {"quote": "Let me tell you the secret that has led me to my goals: my strength lies solely in my tenacity.", "author": "Louis Pasteur"},
    {"quote": "Your employees come first. And if you treat your employees right, guess what? Your customers come back.", "author": "Herb Kelleher"},
    {"quote": "Success isn’t determined by how many times you win, but by how you play the week after you lose.", "author": "Pele"},
    {"quote": "I am happy because I’m grateful. I choose to be grateful. That gratitude allows me to be happy.", "author": "Will Arnett"},
    {"quote": "The true secret of happiness lies in taking a genuine interest in all the details of daily life.", "author": "William Morris"},
    {"quote": "I would rather die of passion than of boredom.", "author": "Vincent van Gogh"},
    {"quote": "The impossible exists only until we find a way to make it possible.", "author": "Mike Horn"},
    {"quote": "Boredom is a lack of crazy. It’s a lack of creativity. Invention. Innovation. If you’re bored, blame yourself.", "author": "Katelyn S. Irons"},
    {"quote": "The greatest danger in times of turbulence is not the turbulence; it is to act with yesterday’s logic.", "author": "Peter Drucker"},
    {"quote": "Never let anyone define you. You are the only person who defines you. No one can speak for you. Only you speak for you. You are your only voice.", "author": "Terry Crews"},
    {"quote": "Be so good they can’t ignore you.", "author": "Steve Martin"},
    {"quote": "Nothing in all the world is more dangerous than sincere ignorance and conscientious stupidity.", "author": "Martin Luther King Jr."},
    {"quote": "The only way to win is to learn faster than anyone else.", "author": "Eric Ries"},
    {"quote": "God doesn’t require us to succeed, he only requires that you try.", "author": "Mother Teresa"},
    {"quote": "May I never be complete. May I never be content. May I never be perfect.", "author": "Chuck Palahniuk"},
    {"quote": "Instead of wondering when your next vacation is, maybe you should set up a life you don’t need to escape from.", "author": "Seth Godin"},
    {"quote": "The greatest test of courage on earth is to bear defeat without losing heart.", "author": "Robert Green Ingersoll"},
    {"quote": "It isn’t hard to be good from time to time in sports. What is tough, is being good every day.", "author": "Willie Mays"},
    {"quote": "Death is very likely the single best invention of life. It’s life’s change agent; it clears out the old to make way for the new. Right now, the new is you. But someday, not too long from now, you will gradually become the old and be cleared away. Sorry to be so dramatic, but it’s quite true. Your time is limited, so don’t waste it living someone else’s life.", "author": "Steve Jobs"},
    {"quote": "Have no fear of perfection. You’ll never reach it.", "author": "Salvador Dalí"},
    {"quote": "The most common way people give up their power is by thinking they don’t have any.", "author": "Alice Walker"},
    {"quote": "The person who agrees with you 80 percent of the time is an 80 percent friend, not a 20 percent enemy.", "author": "Ronald Reagan"},
    {"quote": "The thing about smart people is that they seem like crazy people to dumb people.", "author": "Stephen Hawking"},
    {"quote": "People have confused playing devil’s advocate with being intelligent.", "author": "Cecily Strong"},
    {"quote": "Winning is not everything, but wanting to win is.", "author": "Vince Lombardi"},
    {"quote": "The minute you start caring about what other people think, is the minute you stop being yourself.", "author": "Meryl Streep"},
    {"quote": "Optimism is the one quality more associated with success and happiness than any other.", "author": "Brian Tracy"},
    {"quote": "People who say it cannot be done should not interrupt those who are doing it.", "author": "George Bernard Shaw"},
    {"quote": "My only goal is to stay focused on my craft and make sure my life is as sharp as it can be to attack any character that is given to me.", "author": "Michael K. Williams"},
    {"quote": "If you’re ever thinking, Oh, but I’m a waste of space and I’m a burden, remember: that also describes the Grand Canyon. Why don’t you have friends and family take pictures of you from a safe distance? Revel in your majestic profile?", "author": "Maria Bamford"},
    {"quote": "If everybody loves you, something is wrong. Find at least one enemy to keep you alert.", "author": "Paulo Coelho"},
    {"quote": "The world is more malleable than you think and it’s waiting for you to hammer it into shape.", "author": "Bono"},
    {"quote": "Yes, I am a dreamer. For a dreamer is one who can find his way by moonlight, and see the dawn before the rest of the world.", "author": "Oscar Wilde"},
    {"quote": "Always make a total effort, even when the odds are against you.", "author": "Arnold Palmer"},
    {"quote": "Life doesn’t run away from nobody. Life runs at people.", "author": "Joe Frazier"},
    {"quote": "The last 10 percent it takes to launch something takes as much energy as the first 90 percent.", "author": "Rob Kalin"},
    {"quote": "Don’t let life discourage you; everyone who got where he is had to begin where he was.", "author": "Richard L. Evans"},
    {"quote": "Greatness is a lot of small things done well. Day after day, workout after workout, obedience after obedience, day after day.", "author": "Ray Lewis"},
    {"quote": "Risk more than others think is safe. Dream more than others think is practical.", "author": "Howard Schultz"},
    {"quote": "Next time you think you are important, try ordering somebody else’s dog around.", "author": "Will Rogers"},
    {"quote": "I made a resolve then that I was going to amount to something if I could. And no hours, nor amount of labor, nor amount of money would deter me from giving the best that there was in me. And I have done that ever since, and I win by it. I know.", "author": "Colonel Harland Sanders"},
    {"quote": "If you see a bandwagon, it’s too late.", "author": "James Goldsmith"},
    {"quote": "Everyone has a plan ’til they get punched in the mouth.", "author": "Mike Tyson"},
    {"quote": "I’ve missed more than 9000 shots in my career. I’ve lost almost 300 games. 26 times, I’ve been trusted to take the game winning shot and missed. I’ve failed over and over and over again in my life. And that is why I succeed.", "author": "Michael Jordan"},
    {"quote": "One, don’t believe everything that’s written about you. Two, don’t pick up too many checks.", "author": "Babe Ruth"},
    {"quote": "Before you speak, run this through your head: Is what I’m about to say true? Is it helpful? Is it inspiring? Is it necessary? Is it kind? If you cannot answer yes to these questions, then don’t say it, don’t tweet it, don’t write it.", "author": "Beret Guidera"},
    {"quote": "You can and should set your own limits and clearly articulate them. This takes courage, but it is also liberating and empowering, and often earns you new respect.", "author": "Rosalind Brewer"},
    {"quote": "Every vision is a joke until the first man accomplishes it; once realized, it becomes commonplace.", "author": "Robert Goddard"},
    {"quote": "Don’t be afraid your life will end; be afraid that it will never begin.", "author": "Grace Hansen"},
    {"quote": "When you meet somebody for the first time, you’re not meeting them. You’re meeting their representative.", "author": "Chris Rock"},
    {"quote": "Not everything that can be counted counts, and not everything that counts can be counted.", "author": "William Bruce Cameron"},
    {"quote": "We always honor our people when they die; we’ve got to honor them while we’re still alive.", "author": "Biz Markie"},
    {"quote": "Live your life as an exclamation rather than an explanation.", "author": "Isaac Newton"},
    {"quote": "It’s easy to do nothing, but your heart breaks a little more every time you do.", "author": "Mark Ruffalo"},
    {"quote": "Pressure is a privilege. It only comes to those who earn it.", "author": "Billie Jean King"},
    {"quote": "There will come a time when you believe everything is finished. That will be the beginning.", "author": "Louis L’Amour"},
    {"quote": "The truth is, everyone is going to hurt you. You just got to find the ones worth suffering for.", "author": "Bob Marley"},
    {"quote": "Give me a stock clerk with a goal, and I will give you a man who will make history. Give me a man without a goal, and I will give you a stock clerk.", "author": "James Cash Penney"},
    {"quote": "I shall never pay a dollar of your unjust penalty.", "author": "Susan B. Anthony"},
    {"quote": "My first job was washing dishes in the basement of a nursing home for $2.10 an hour, and I learned as much about the value of hard work there as I ever did later.", "author": "Douglas Preston"},
    {"quote": "Understand well as I may, my comprehension can only be an infinitesimal fraction of all I want to understand.", "author": "Ada Lovelace"},
    {"quote": "When you are not willing to fully receive, you are training the universe not to give to you! It’s simple: if you aren’t willing to receive your share, it will go to someone else who is.", "author": "T. Harv Eker"},
    {"quote": "You’re going to fall down, but the world doesn’t care how many times you fall down, as long as it’s one fewer than the numbers of times you get back up.", "author": "Aaron Sorkin"},
    {"quote": "You do this because you like it, you think what you’re making is beautiful. And if you think it’s beautiful, maybe they think it’s beautiful.", "author": "Lou Reed"},
    {"quote": "We are less bored than our ancestors were, but we are more afraid of boredom.", "author": "Bertrand Russell"},
    {"quote": "Luck? I don’t know anything about luck. I’ve never banked on it, and I’m afraid of people who do. Luck to me is something else: hard work, and realizing what is opportunity and what isn’t.", "author": "Lucille Ball"},
    {"quote": "Don’t aspire to make a living, aspire to make a difference.", "author": "Denzel Washington"},
    {"quote": "It’s good to learn from your mistakes. It’s better to learn from other people’s mistakes.", "author": "Warren Buffett"},
    {"quote": "Every man dies. Not every man really lives.", "author": "William Wallace"},
    {"quote": "I feel like if it’s not scaring you, you’re doing it wrong.", "author": "Anna Kendrick"},
    {"quote": "You don’t become what you want, you become what you believe.", "author": "Oprah Winfrey"},
    {"quote": "There is no element of genius without some form of madness.", "author": "Aristotle"},
    {"quote": "Sometimes you have to wander a bit, and do what you don’t want to in order to figure out what it is you’re supposed to do.", "author": "Larry David"},
    {"quote": "Life always offers you a second chance. It is called tomorrow.", "author": "Dylan Thomas"},
    {"quote": "To scale, do things that don’t scale.", "author": "Reid Hoffman"},
    {"quote": "Normal is not something to aspire to, it’s something to get away from.", "author": "Jodie Foster"},
    {"quote": "Sometimes your joy is the source of your smile, but sometimes your smile can be the source of your joy.", "author": "Thich Nhat Hanh"},
    {"quote": "Men give me credit for some genius. All the genius I have lies in this; when I have a subject in hand, I study it profoundly. Day and night it is before me. My mind becomes pervaded with it. Then the effort that I have made is what people are pleased to call the fruit of genius. It is the fruit of labor and thought.", "author": "Alexander Hamilton"},
    {"quote": "Time is life itself, and life resides in the human heart.", "author": "Michael Ende"},
    {"quote": "The big money is not in the buying and selling … but in the waiting.", "author": "Charlie Munger"},
    {"quote": "Gratitude is not only the greatest of virtues, but the parent of all the others.", "author": "Marcus Tullius Cicero"},
    {"quote": "Someone once said, 'Adversity introduces a man to himself.' For some reason, that’s scary, but most people discover that adversity does make them stronger.", "author": "Max Cleland"},
    {"quote": "I can’t tell you which one is right. But I can tell you which one is more fun.", "author": "Phil Knight"},
    {"quote": "The lessons aren’t about wealth or fame or working harder and harder. The clearest message that we get from this 75-year study is this: Good relationships keep us happier and healthier. Period.", "author": "Robert Waldinger"},
    {"quote": "I put instant coffee in a microwave oven and almost went back in time.", "author": "Steven Wright"},
    {"quote": "The best way out is always through.", "author": "Robert Frost"},
    {"quote": "If you want to lift yourself up, lift up someone else.", "author": "Booker T. Washington"},
    {"quote": "Still and all, why bother? Here’s my answer. …", "author": "Kurt Vonnegut"},
    {"quote": "In every good marriage, it helps sometimes to be a little deaf.", "author": "Ruth Bader Ginsburg"},
    {"quote": "People think I’m an overnight success. No. It’s just that you all found me overnight.", "author": "Leslie Jones"},
    {"quote": "To bring about change, you must not be afraid to take the first step. We will fail when we fail to try.", "author": "Rosa Parks"},
    {"quote": "If you think you are too small to make a difference, try sleeping with a mosquito.", "author": "Dalai Lama"},
    {"quote": "There is only one boss: the customer. And he can fire everybody in the company, from the chairman on down, simply by spending his money somewhere else.", "author": "Sam Walton"},
    {"quote": "I always want to say to people who want to be rich and famous: 'Try being rich first.' See if that doesn’t cover most of it.", "author": "Bill Murray"},
    {"quote": "We don’t want to suffer. We don’t want to feel discomfort. So the whole time, we’re living our lives in a very comfortable area. There’s no growth in that.", "author": "David Goggins"},
    {"quote": "It’s better for people to miss you than to have seen too much of you.", "author": "Edward Norton"},
    {"quote": "Forget about style; worry about results.", "author": "Bobby Orr"},
    {"quote": "St. Francis of Assisi taught me that there is a wound in the Creation and that the greatest use we could make of our lives was to ask to be made a healer of it.", "author": "Alan Paton"},
    {"quote": "My motto was always to keep swinging. Whether I was in a slump or feeling badly or having trouble off the field, the only thing to do was keep swinging.", "author": "Hank Aaron"},
    {"quote": "Developing a good work ethic is key. Apply yourself at whatever you do, whether you’re a janitor or taking your first summer job, because that work ethic will be reflected in everything you do in life.", "author": "Tyler Perry"},
    {"quote": "The road to success is not easy to navigate, but with hard work, drive, and passion, it’s possible to achieve the American dream.", "author": "Tommy Hilfiger"},
    {"quote": "I don’t like to gamble, but if there’s one thing I’m willing to bet on, it’s myself.", "author": "Beyoncé"},
    {"quote": "Nobody in life gets exactly what they thought they were going to get. But if you work really hard and you’re kind, amazing things will happen.", "author": "Conan O’Brien"},
    {"quote": "Pray as though everything depended on God. Work as though everything depended on you.", "author": "Saint Augustine"},
    {"quote": "If you don’t fall how are you going to know what getting up is like.", "author": "Stephen Curry"},
    {"quote": "Frankly, it’s time for a new generation of leaders.", "author": "Mitt Romney"},
    {"quote": "To be a Christian means to forgive the inexcusable because God has forgiven the inexcusable in you.", "author": "C. S. Lewis"},
    {"quote": "I always prefer to believe the best of everybody, it saves so much trouble.", "author": "Rudyard Kipling"},
    {"quote": "This is it. This is life, the one you get.", "author": "Jeff Barry"},
    {"quote": "Greatness is not measured by what you build, but by what you inspire others to create.","author": "Raul M. Uñate"}
]



from orionis.foundation.config.testing.enums.verbosity import VerbosityMode
from orionis.test.exceptions import OrionisTestValueError

class __ValidVerbosity:
    """
    Validator for verbosity levels, ensuring the input is a non-negative integer corresponding to a valid
    VerbosityMode value or an instance of VerbosityMode.

    This class is intended to validate verbosity arguments for test configuration.
    """

    def __call__(self, verbosity) -> int:
        """
        Validate the verbosity level.

        Parameters
        ----------
        verbosity : int or VerbosityMode
            The verbosity level to validate. Must be a non-negative integer matching a VerbosityMode value,
            or an instance of VerbosityMode.

        Returns
        -------
        int
            The validated verbosity level as an integer.

        Raises
        ------
        OrionisTestValueError
            If the verbosity is not a non-negative integer corresponding to a VerbosityMode value,
            nor an instance of VerbosityMode.
        """
        if isinstance(verbosity, VerbosityMode):
            return verbosity.value
        if isinstance(verbosity, int) and verbosity >= 0:
            if verbosity in [mode.value for mode in VerbosityMode]:
                return verbosity
            else:
                raise OrionisTestValueError(
                    f"Invalid verbosity level: {verbosity} is not a valid VerbosityMode value."
                )
        raise OrionisTestValueError(
            f"Invalid verbosity level: Expected a non-negative integer or VerbosityMode, got '{verbosity}' ({type(verbosity).__name__})."
        )

# Exported singleton instance
ValidVerbosity = __ValidVerbosity()


import asyncio
from datetime import datetime
import logging
from typing import Any, Dict, List, Optional, Tuple, Union
from apscheduler.events import (
    EVENT_JOB_ERROR,
    EVENT_JOB_EXECUTED,
    EVENT_JOB_MAX_INSTANCES,
    EVENT_JOB_MISSED,
    EVENT_JOB_REMOVED,
    EVENT_JOB_SUBMITTED,
    EVENT_SCHEDULER_PAUSED,
    EVENT_SCHEDULER_RESUMED,
    EVENT_SCHEDULER_SHUTDOWN,
    EVENT_SCHEDULER_STARTED,
)
from apscheduler.schedulers.asyncio import AsyncIOScheduler as APSAsyncIOScheduler
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.schedulers.base import STATE_PAUSED, STATE_RUNNING
from apscheduler.triggers.date import DateTrigger
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from orionis.console.contracts.event import IEvent
from orionis.console.contracts.reactor import IReactor
from orionis.console.contracts.schedule import ISchedule
from orionis.console.contracts.schedule_event_listener import IScheduleEventListener
from orionis.console.entities.event_job import EventJob
from orionis.console.entities.scheduler_error import SchedulerError
from orionis.console.entities.scheduler_paused import SchedulerPaused
from orionis.console.entities.scheduler_resumed import SchedulerResumed
from orionis.console.entities.scheduler_shutdown import SchedulerShutdown
from orionis.console.entities.scheduler_started import SchedulerStarted
from orionis.console.entities.event import Event as EventEntity
from orionis.console.enums.listener import ListeningEvent
from orionis.console.exceptions import CLIOrionisRuntimeError
from orionis.console.exceptions import CLIOrionisValueError
from orionis.console.request.cli_request import CLIRequest
from orionis.failure.contracts.catch import ICatch
from orionis.foundation.contracts.application import IApplication
from orionis.services.log.contracts.log_service import ILogger
from zoneinfo import ZoneInfo

class Schedule(ISchedule):

    # Error message constants
    SIGNATURE_REQUIRED_ERROR = "Signature must be a non-empty string."
    NOT_APPLICABLE = "Not Applicable"

    def __init__(
        self,
        reactor: IReactor,
        app: IApplication,
        rich_console: Console
    ) -> None:
        """
        Initialize a new instance of the Schedule class.

        This constructor sets up the internal state required for scheduling commands,
        including references to the application instance, the AsyncIOScheduler, the
        command reactor, and job tracking structures. It also initializes properties
        for managing the current scheduling context, logging, and event listeners.

        Parameters
        ----------
        reactor : IReactor
            An instance implementing the IReactor interface, used to retrieve available
            commands and execute scheduled jobs.
        app : IApplication
            The application container instance, used for configuration, dependency injection,
            and service resolution.
        rich_console : Console
            An instance of Rich's Console for advanced output formatting.

        Returns
        -------
        None
            This method does not return any value. It initializes the Schedule instance
            and prepares all required internal structures for scheduling and event handling.
        """

        # List of operations that can be performed on the scheduler
        self.__control_operations = [
            'schedule:pause',
            'schedule:resume',
            'schedule:shutdown'
        ]

        # Store the application instance for configuration and service access.
        self.__app: IApplication = app

        # Store the rich console instance for advanced output formatting.
        self.__rich_console = rich_console

        # Save the timezone configuration from the application settings.
        self.__tz = ZoneInfo(self.__app.config('app.timezone') or 'UTC')

        # Initialize the AsyncIOScheduler with the configured timezone.
        self.__scheduler = APSAsyncIOScheduler(timezone=self.__tz)
        self.__control_scheduler = APSAsyncIOScheduler(timezone=self.__tz)

        # Disable APScheduler's internal logging to avoid duplicate/conflicting logs.
        self.__disableLogging()

        # Initialize the logger using the application's dependency injection.
        self.__logger: ILogger = self.__app.make(ILogger)

        # Store the reactor instance for command management and job execution.
        self.__reactor: IReactor = reactor

        # Retrieve and store all available commands from the reactor.
        self.__available_commands = self.__getCommands()

        # Initialize the dictionary to keep track of scheduled events.
        self.__events: Dict[str, IEvent] = {}

        # Initialize the list to keep track of all scheduled job entities.
        self.__jobs: List[EventEntity] = []

        # Initialize the dictionary to manage event listeners.
        self.__listeners: Dict[str, callable] = {}

        # Initialize the asyncio event used to signal scheduler shutdown.
        self._stopEvent: Optional[asyncio.Event] = None

        # Retrieve and initialize the catch instance for exception handling.
        self.__catch: ICatch = app.make(ICatch)

        # Initialize the paused at timestamp
        self.__paused_at: Optional[datetime] = None

    def __disableLogging(
        self
    ) -> None:
        """
        Disable APScheduler logging to prevent conflicts and duplicate log messages.

        This method disables logging for the APScheduler library and its key subcomponents.
        It clears all handlers attached to the APScheduler loggers, disables log propagation,
        and turns off the loggers entirely. This is useful in applications that have their
        own logging configuration and want to avoid duplicate or unwanted log output from
        APScheduler.

        Returns
        -------
        None
            This method does not return any value. It modifies the logging configuration
            of APScheduler loggers in place.
        """

        # List of APScheduler logger names to disable
        for name in ["apscheduler", "apscheduler.scheduler", "apscheduler.executors.default"]:
            logger = logging.getLogger(name)
            logger.handlers.clear()                 # Remove all handlers to prevent output
            logger.propagate = False                # Prevent log messages from propagating to ancestor loggers
            logger.disabled = True                  # Disable the logger entirely

    def __getCommands(
        self
    ) -> dict:
        """
        Retrieve available commands from the reactor and return them as a dictionary.

        This method queries the reactor for all available jobs/commands, extracting their
        signatures and descriptions. The result is a dictionary where each key is the command
        signature and the value is another dictionary containing the command's signature and
        its description.

        Returns
        -------
        dict
            A dictionary mapping command signatures to their details. Each value is a dictionary
            with 'signature' and 'description' keys.
        """

        # Initialize the commands dictionary
        commands = {}

        # Iterate over all jobs provided by the reactor's info method
        for job in self.__reactor.info():

            signature: str = job.get('signature', None)
            description: str = job.get('description', 'No description available.')

            # Skip invalid or special method signatures
            if not signature or (signature.startswith('__') and signature.endswith('__')):
                continue

            # Store each job's signature and description in the commands dictionary
            commands[signature] = {
                'signature': signature,
                'description': description
            }

        # Return the commands dictionary
        return commands

    def __getCurrentTime(
        self
    ) -> str:
        """
        Retrieve the current date and time as a formatted string in the configured timezone.

        This method obtains the current date and time using the timezone specified in the application's
        configuration (defaulting to UTC if not set). The result is formatted as a string in the
        "YYYY-MM-DD HH:MM:SS" format, which is suitable for logging, display, or timestamping events.

        Returns
        -------
        str
            The current date and time as a string in the format "YYYY-MM-DD HH:MM:SS", localized to the
            scheduler's configured timezone.
        """

        # Get the current time in the scheduler's configured timezone
        now = datetime.now(self.__tz)

        # Return the formatted current time string
        return now.strftime("%Y-%m-%d %H:%M:%S")

    def __getNow(
        self
    ) -> datetime:
        """
        Retrieve the current date and time as a timezone-aware datetime object.

        This method returns the current date and time, localized to the timezone configured
        for the scheduler instance. The timezone is determined by the application's configuration
        (typically set during initialization). This is useful for ensuring that all time-related
        operations within the scheduler are consistent and respect the application's timezone settings.

        Returns
        -------
        datetime
            A timezone-aware `datetime` object representing the current date and time
            in the scheduler's configured timezone.
        """

        # Return the current datetime localized to the scheduler's timezone
        return datetime.now(self.__tz)

    def __getAttribute(
        self,
        obj: Any,
        attr: str,
        default: Any = None
    ) -> Any:
        """
        Safely retrieve an attribute from an object, returning a default value if the attribute does not exist.

        This method attempts to access the specified attribute of the given object. If the attribute
        exists, its value is returned. If the attribute does not exist, the provided default value
        is returned instead. This prevents `AttributeError` exceptions when accessing attributes
        that may not be present on the object.

        Parameters
        ----------
        obj : Any
            The object from which to retrieve the attribute.
        attr : str
            The name of the attribute to retrieve.
        default : Any, optional
            The value to return if the attribute does not exist. Defaults to None.

        Returns
        -------
        Any
            The value of the specified attribute if it exists on the object; otherwise, the value
            of `default`.

        Notes
        -----
        This method is a wrapper around Python's built-in `getattr()` function, providing a safe
        way to access attributes that may or may not exist on an object.
        """

        # If the object is None, return the default value immediately
        if obj is None:
            return default

        # Use Python's built-in getattr to safely retrieve the attribute,
        # returning the default value if the attribute is not found.
        return getattr(obj, attr, default)

    def __isAvailable(
        self,
        signature: str
    ) -> bool:
        """
        Check if a command with the given signature is available among registered commands.

        This method determines whether the provided command signature exists in the internal
        dictionary of available commands. It is used to validate if a command can be scheduled
        or executed by the scheduler.

        Parameters
        ----------
        signature : str
            The signature of the command to check for availability.

        Returns
        -------
        bool
            Returns True if the command with the specified signature exists in the available
            commands dictionary; otherwise, returns False.

        Notes
        -----
        The method performs a simple membership check in the internal commands dictionary.
        It does not validate the format or correctness of the signature itself, only its
        presence among the registered commands.
        """

        # Check if the signature exists in the available commands dictionary
        return signature in self.__available_commands

    def __getDescription(
        self,
        signature: str
    ) -> Optional[str]:
        """
        Retrieve the description for a command based on its signature.

        This method searches the internal dictionary of available commands for the provided
        command signature and returns the corresponding description if found. If the signature
        does not exist in the available commands, the method returns None.

        Parameters
        ----------
        signature : str
            The unique signature identifying the command whose description is to be retrieved.

        Returns
        -------
        Optional[str]
            The description string associated with the command signature if it exists;
            otherwise, returns None.

        Notes
        -----
        This method is useful for displaying human-readable information about commands
        when scheduling or listing tasks.
        """

        # Attempt to retrieve the command entry from the available commands dictionary
        command_entry = self.__available_commands.get(signature)

        # If the command entry exists, return its description; otherwise, return None
        return command_entry['description'] if command_entry else None

    def command(
        self,
        signature: str,
        args: Optional[List[str]] = None
    ) -> 'IEvent':
        """
        Prepare and register an Event instance for a given command signature and its arguments.

        This method validates the provided command signature and arguments, ensuring that the command
        exists among the registered commands and that the arguments are in the correct format. If validation
        passes, it creates and registers an Event object representing the scheduled command, including its
        signature, arguments, and description. The Event instance is stored internally for later scheduling.

        Parameters
        ----------
        signature : str
            The unique signature identifying the command to be scheduled. Must be a non-empty string.
        args : Optional[List[str]], optional
            A list of string arguments to be passed to the command. Defaults to None.

        Returns
        -------
        IEvent
            The Event instance containing the command signature, arguments, and its description.
            This instance is also stored internally for further scheduling configuration.

        Raises
        ------
        CLIOrionisValueError
            If the command signature is not a non-empty string, if the arguments are not a list
            of strings or None, or if the command does not exist among the registered commands.
        """

        # Prevent adding new commands while the scheduler is running
        if self.isRunning():
            self.__raiseException(
                CLIOrionisValueError(
                    "Cannot add new commands while the scheduler is running. Please stop the scheduler before adding new commands."
                )
            )

        # Validate that the command signature is a non-empty string
        if not isinstance(signature, str) or not signature.strip():
            raise CLIOrionisValueError(
                "The command signature must be a non-empty string. Please provide a valid command signature."
            )

        # Ensure that arguments are either a list of strings or None
        if args is not None:
            if not isinstance(args, list):
                raise CLIOrionisValueError(
                    "Arguments must be provided as a list of strings or None. Please check your arguments."
                )
            for arg in args:
                if not isinstance(arg, str):
                    raise CLIOrionisValueError(
                        f"Invalid argument '{arg}'. Each argument must be a string."
                    )

        # Check if the command is available in the registered commands
        if not self.__isAvailable(signature):
            raise CLIOrionisValueError(
                f"The command '{signature}' is not available or does not exist. Please check the command signature."
            )

        # Import Event here to avoid circular dependency issues
        from orionis.console.fluent.event import Event

        # Store the command and its arguments for scheduling in the internal events dictionary
        self.__events[signature] = Event(
            signature=signature,
            args=args or [],
            purpose=self.__getDescription(signature)
        )

        # Return the Event instance for further scheduling configuration
        return self.__events[signature]

    def __getTaskFromSchedulerById(
        self,
        id: str,
        code: int = None
    ) -> Optional[EventJob]:
        """
        Retrieve a scheduled job from the AsyncIOScheduler by its unique ID.

        This method fetches a job from the AsyncIOScheduler using its unique identifier (ID).
        It extracts the job's attributes and creates a `Job` entity containing the relevant
        details. If the job does not exist, the method returns `None`.

        Parameters
        ----------
        id : str
            The unique identifier (ID) of the job to retrieve. This must be a non-empty string.

        Returns
        -------
        Optional[Job]
            A `Job` entity containing the details of the scheduled job if it exists.
            If the job does not exist, returns `None`.
        """

        # Extract event data from the internal events list if available
        event_data: dict = self.event(id)

        # Retrieve the job data from the scheduler using the provided job ID
        data = self.__scheduler.get_job(id)

        # If no job is found, return EventJob with default values
        _id = self.__getAttribute(data, 'id', None)
        if not _id and code in (EVENT_JOB_MISSED, EVENT_JOB_REMOVED):
            _id = event_data.get('signature')
        elif not _id:
            return EventJob()

        # Extract the job code if available
        _code = code if code is not None else 0

        # Extract the job name if available
        _name = self.__getAttribute(data, 'name', None)

        # Extract the job function if available
        _func = self.__getAttribute(data, 'func', None)

        # Extract the job arguments if available
        _args = self.__getAttribute(data, 'args', tuple(event_data.get('args', [])))

        # Extract the job trigger if available
        _trigger = self.__getAttribute(data, 'trigger', None)

        # Extract the job executor if available
        _executor = self.__getAttribute(data, 'executor', None)

        # Extract the job jobstore if available
        _jobstore = self.__getAttribute(data, 'jobstore', None)

        # Extract the job misfire_grace_time if available
        _misfire_grace_time = self.__getAttribute(data, 'misfire_grace_time', None)

        # Extract the job max_instances if available
        _max_instances = self.__getAttribute(data, 'max_instances', 1)

        # Extract the job coalesce if available
        _coalesce = self.__getAttribute(data, 'coalesce', False)

        # Extract the job next_run_time if available
        _next_run_time = self.__getAttribute(data, 'next_run_time', None)

        # Extract additional event data if available
        _purpose = event_data.get('purpose', None)

        # Extract additional event data if available
        _start_date = event_data.get('start_date', None)

        # Extract additional event data if available
        _end_date = event_data.get('end_date', None)

        # Extract additional event data if available
        _details = event_data.get('details', None)

        # Create and return a Job entity based on the retrieved job data
        return EventJob(
            id=_id,
            code=_code,
            name=_name,
            func=_func,
            args=_args,
            trigger=_trigger,
            executor=_executor,
            jobstore=_jobstore,
            misfire_grace_time=_misfire_grace_time,
            max_instances=_max_instances,
            coalesce=_coalesce,
            next_run_time=_next_run_time,
            exception = None,
            traceback = None,
            retval = None,
            purpose = _purpose,
            start_date = _start_date,
            end_date = _end_date,
            details = _details
        )

    def __subscribeListeners(
        self
    ) -> None:
        """
        Subscribe internal handlers to APScheduler events for monitoring and control.

        This method attaches internal listener methods to the AsyncIOScheduler instance for a variety of
        scheduler and job-related events. These listeners enable the scheduler to respond to lifecycle
        changes (such as start, shutdown, pause, and resume) and job execution events (such as submission,
        execution, errors, missed runs, max instance violations, and removals).

        Each listener is responsible for logging, invoking registered callbacks, and handling errors or
        missed jobs as appropriate. This setup ensures that the scheduler's state and job execution
        are monitored and managed effectively.

        Returns
        -------
        None
            This method does not return any value. It registers event listeners on the scheduler instance.

        Notes
        -----
        The listeners are attached to the following APScheduler events:
            - Scheduler started
            - Scheduler shutdown
            - Job error
            - Job submitted
            - Job executed
            - Job missed
            - Job max instances exceeded
            - Job removed

        These listeners enable the scheduler to handle and react to all critical scheduling events.
        """

        # Register listener for when the scheduler starts
        self.__scheduler.add_listener(self.__startedListener, EVENT_SCHEDULER_STARTED)

        # Register listener for when the scheduler shuts down
        self.__scheduler.add_listener(self.__shutdownListener, EVENT_SCHEDULER_SHUTDOWN)

        # Register listener for job execution errors
        self.__scheduler.add_listener(self.__errorListener, EVENT_JOB_ERROR)

        # Register listener for when a job is submitted to its executor
        self.__scheduler.add_listener(self.__submittedListener, EVENT_JOB_SUBMITTED)

        # Register listener for when a job has finished executing
        self.__scheduler.add_listener(self.__executedListener, EVENT_JOB_EXECUTED)

        # Register listener for when a job is missed (not run at its scheduled time)
        self.__scheduler.add_listener(self.__missedListener, EVENT_JOB_MISSED)

        # Register listener for when a job exceeds its maximum allowed concurrent instances
        self.__scheduler.add_listener(self.__maxInstancesListener, EVENT_JOB_MAX_INSTANCES)

        # Register listener for when a job is removed from the scheduler
        self.__scheduler.add_listener(self.__removedListener, EVENT_JOB_REMOVED)

    def __globalCallableListener(
        self,
        event_data: Optional[Union[SchedulerStarted, SchedulerPaused, SchedulerResumed, SchedulerShutdown, SchedulerError]],
        listening_vent: ListeningEvent
    ) -> None:
        """
        Invoke registered listeners for global scheduler events.

        This method is responsible for handling global scheduler events such as when the scheduler starts, pauses, resumes,
        shuts down, or encounters an error. It checks if a listener is registered for the specified event type and, if so,
        invokes the listener with the provided event data and the current scheduler instance. The listener can be either a
        coroutine or a regular callable.

        Parameters
        ----------
        event_data : Optional[Union[SchedulerStarted, SchedulerPaused, SchedulerResumed, SchedulerShutdown, SchedulerError]]
            The event data associated with the global scheduler event. This may include details about the event,
            such as its type, code, time, or context. If no specific data is available, this parameter can be None.
        listening_vent : ListeningEvent
            An instance of the ListeningEvent enum representing the global scheduler event to handle.

        Returns
        -------
        None
            This method does not return any value. It executes the registered listener for the specified event if one exists.

        Raises
        ------
        CLIOrionisValueError
            If the provided `listening_vent` is not an instance of ListeningEvent.

        Notes
        -----
        This method is intended for internal use to centralize the invocation of global event listeners.
        It ensures that only valid event types are processed and that exceptions raised by listeners are
        properly handled through the application's error handling mechanism.
        """

        # Ensure the provided event is a valid ListeningEvent instance
        if not isinstance(listening_vent, ListeningEvent):
            self.__raiseException(CLIOrionisValueError("The event must be an instance of ListeningEvent."))

        # Retrieve the string identifier for the event from the ListeningEvent enum
        scheduler_event = listening_vent.value

        # Check if a listener is registered for this global event
        if scheduler_event in self.__listeners:

            # Retrieve the listener callable for the event
            listener = self.__listeners[scheduler_event]

            # If the listener is callable, invoke it with event data and the scheduler instance
            if callable(listener):

                try:

                    # Execute the listener, handling both coroutine and regular functions
                    self.__app.invoke(listener, event_data, self)

                except Exception as e:

                    # Handle any exceptions raised by the listener using the application's error handler
                    self.__raiseException(e)

    def __taskCallableListener( # NOSONAR
        self,
        event_data: EventJob,
        listening_vent: ListeningEvent
    ) -> None:
        """
        Invoke registered listeners for specific task/job events.

        This method is responsible for handling task/job-specific events such as job errors,
        executions, submissions, missed jobs, and max instance violations. It checks if a
        listener is registered for the specific job ID associated with the event and, if so,
        invokes the appropriate method on the listener. The listener can be either a class
        implementing `IScheduleEventListener` or a callable.

        Parameters
        ----------
        event_data : EventJob
            The event data associated with the task/job event. This includes details about the job,
            such as its ID, exception (if any), and other context. If no specific data is available,
            this parameter can be None.
        listening_vent : ListeningEvent
            An instance of the `ListeningEvent` enum representing the task/job event to handle.

        Returns
        -------
        None
            This method does not return any value. It invokes the registered listener for the
            specified job event, if one exists.

        Raises
        ------
        CLIOrionisValueError
            If the provided `listening_vent` is not an instance of `ListeningEvent`.
            If the listener for the job ID is not a subclass of `IScheduleEventListener`.

        Notes
        -----
        This method is intended for internal use to centralize the invocation of task/job event listeners.
        It ensures that only valid event types and listeners are processed, and that exceptions raised
        by listeners are properly handled through the application's error handling mechanism.
        """

        # Ensure the provided event is a valid ListeningEvent instance
        if not isinstance(listening_vent, ListeningEvent):
            self.__raiseException(CLIOrionisValueError("The event must be an instance of ListeningEvent."))

        # Validate that event_data is a valid EventJob with a non-empty id
        if not isinstance(event_data, EventJob) or not hasattr(event_data, 'id') or not event_data.id:
            return

        # Retrieve the string identifier for the event from the ListeningEvent enum
        scheduler_event = listening_vent.value

        # Check if a listener is registered for this specific job ID
        if event_data.id in self.__listeners:

            # Retrieve the listener for the specific job ID
            listener = self.__listeners[event_data.id]

            # If the listener is a subclass of IScheduleEventListener, invoke the appropriate method
            if issubclass(listener, IScheduleEventListener):
                try:
                    # If the listener is a class, instantiate it using the application container
                    if isinstance(listener, type):
                        listener = self.__app.make(listener)

                    # Check if the listener has a method corresponding to the event type and is callable
                    if hasattr(listener, scheduler_event) and callable(getattr(listener, scheduler_event)):
                        # Call the event method on the listener, passing event data and the scheduler instance
                        self.__app.call(listener, scheduler_event, event_data, self)

                except Exception as e:

                    # Handle any exceptions raised by the listener using the application's error handler
                    self.__raiseException(e)

            else:
                # If the listener is not a subclass of IScheduleEventListener, raise an exception
                self.__raiseException(
                    CLIOrionisValueError(
                        f"The listener for job ID '{event_data.id}' must be a subclass of IScheduleEventListener."
                    )
                )

    def __getPID(
        self
    ) -> int:
        """
        Retrieve the current process ID (PID) of the running scheduler.

        This method obtains the process ID of the current Python process using the `os` module.
        The PID is a unique identifier assigned by the operating system to each running process.
        This information can be useful for logging, monitoring, or managing the scheduler process.

        Returns
        -------
        int
            The process ID (PID) of the current Python process.
        """

        import os

        # Return the current process ID using os.getpid()
        return os.getpid()

    def __startedListener(
        self,
        event
    ) -> None:
        """
        Handle the scheduler started event for logging and invoking registered listeners.

        This method is triggered when the scheduler starts. It logs an informational
        message indicating that the scheduler has started successfully and displays
        a formatted message on the rich console. If a listener is registered for the
        scheduler started event, it invokes the listener with the event details.

        Parameters
        ----------
        event : SchedulerStarted
            An event object containing details about the scheduler start event.

        Returns
        -------
        None
            This method does not return any value. It performs logging, displays
            a message on the console, and invokes any registered listener for the
            scheduler started event.
        """

        # Get the current time in the configured timezone
        now = self.__getCurrentTime()

        # Display a start message for the scheduler worker on the rich console
        # Add a blank line for better formatting
        self.__rich_console.line()
        panel_content = Text.assemble(
            ("🚀 Orionis Scheduler Worker ", "bold white on green"),
            ("\n\n", ""),
            ("✅ The scheduled tasks worker has started successfully.\n", "white"),
            (f"🕒 Started at: {now} | 🌐 Timezone: {self.__tz.key} | 🆔 PID: {self.__getPID()}\n", "dim"),
            ("🛑 To stop the worker, press ", "white"),
            ("Ctrl+C", "bold yellow")
        )

        # Display the message in a styled panel
        self.__rich_console.print(
            Panel(panel_content, border_style="green", padding=(1, 2))
        )

        # Add another blank line for better formatting
        self.__rich_console.line()

        # Check if a listener is registered for the scheduler started event
        event_data = SchedulerStarted(
            code=self.__getAttribute(event, 'code', 0),
            time=self.__getNow()
        )

        # If a listener is registered for this event, invoke the listener with the event details
        self.__globalCallableListener(event_data, ListeningEvent.SCHEDULER_STARTED)

        # Log an informational message indicating that the scheduler has started
        self.__logger.info(f"Orionis Scheduler started successfully at: {now}.")

        # Log the timezone currently assigned to the scheduler for traceability
        self.__logger.info(
            f"Timezone assigned to the scheduler: {self.__tz.key}."
        )

    def __shutdownListener(
        self,
        event
    ) -> None:
        """
        Handle the scheduler shutdown event for logging and invoking registered listeners.

        This method is triggered when the scheduler shuts down. It logs an informational
        message indicating that the scheduler has shut down successfully and displays
        a formatted message on the rich console. If a listener is registered for the
        scheduler shutdown event, it invokes the listener with the event details.

        Parameters
        ----------
        event : SchedulerShutdown
            An event object containing details about the scheduler shutdown event.

        Returns
        -------
        None
            This method does not return any value. It performs logging, displays
            a message on the console, and invokes any registered listener for the
            scheduler shutdown event.
        """

        # Get the current time in the configured timezone
        now = self.__getCurrentTime()

        # Check if a listener is registered for the scheduler shutdown event
        event_data = SchedulerShutdown(
            code=self.__getAttribute(event, 'code', 0),
            time=self.__getNow()
        )

        # If a listener is registered for this event, invoke the listener with the event details
        self.__globalCallableListener(event_data, ListeningEvent.SCHEDULER_SHUTDOWN)

        # Log an informational message indicating that the scheduler has shut down
        self.__logger.info(f"Orionis Scheduler shut down successfully at {now}.")

    def __errorListener(
        self,
        event
    ) -> None:
        """
        Handle job error events for logging, error reporting, and listener invocation.

        This method is triggered when a scheduled job raises an exception during execution.
        It logs an error message with the job ID and exception details, updates the job event
        data with error information, and invokes any registered listeners for both the specific
        job error and the global scheduler error event. The method also delegates exception
        handling to the application's error catching mechanism.

        Parameters
        ----------
        event : JobError
            An event object containing details about the errored job, including its ID,
            exception, traceback, and event code.

        Returns
        -------
        None
            This method does not return any value. It performs logging, error reporting,
            and invokes any registered listeners for the job error and scheduler error events.

        Notes
        -----
        - The method updates the job event data with the exception and traceback.
        - Both job-specific and global error listeners are invoked if registered.
        - All exceptions are delegated to the application's error handling system.
        """

        # Extract job ID, event code, exception, and traceback from the event object
        event_id = self.__getAttribute(event, 'job_id', None)
        event_code = self.__getAttribute(event, 'code', 0)
        event_exception = self.__getAttribute(event, 'exception', None)
        event_traceback = self.__getAttribute(event, 'traceback', None)

        # Retrieve the job event data and update it with error details
        job_event_data = self.__getTaskFromSchedulerById(event_id)
        job_event_data.code = event_code
        job_event_data.exception = event_exception
        job_event_data.traceback = event_traceback

        # Invoke the task-specific listener for job errors, if registered
        self.__taskCallableListener(job_event_data, ListeningEvent.JOB_ON_FAILURE)

        # Prepare the global scheduler error event data
        event_data = SchedulerError(
            code=event_code,
            time=self.__getNow(),
            exception=event_exception,
            traceback=event_traceback
        )

        # Invoke the global listener for scheduler errors, if registered
        self.__globalCallableListener(event_data, ListeningEvent.SCHEDULER_ERROR)

        # Delegate exception handling to the application's error catching mechanism
        self.__raiseException(event_exception)

    def __submittedListener(
        self,
        event
    ) -> None:
        """
        Handle job submission events for logging and invoking registered listeners.

        This internal method is triggered when a job is submitted to its executor by the scheduler.
        It logs an informational message about the job submission, creates an event entity representing
        the job submission, and invokes any registered listeners for the submitted job. This allows
        for custom pre-execution logic or notifications to be handled externally.

        Parameters
        ----------
        event : JobSubmitted
            An event object containing details about the submitted job, such as its ID and
            any associated event code.

        Returns
        -------
        None
            This method does not return any value. It performs logging and invokes any
            registered listener for the job submission event.

        Notes
        -----
        This method is intended for internal use to centralize the handling of job submission
        events. It ensures that job submissions are logged and that any custom listeners
        associated with the job are properly notified.
        """

        # Extract job ID and code from the event object, using default values if not present
        event_id = self.__getAttribute(event, 'job_id', None)
        event_code = self.__getAttribute(event, 'code', 0)

        # Create an event entity for the submitted job, including its ID and code
        data_event = self.__getTaskFromSchedulerById(event_id, event_code)

        # If a listener is registered for this job ID, invoke the listener with the event details
        self.__taskCallableListener(data_event, ListeningEvent.JOB_BEFORE)

    def __executedListener(
        self,
        event
    ) -> None:
        """
        Handle job execution events for logging, error reporting, and listener invocation.

        This method is triggered when a job has finished executing in the scheduler. It logs an informational
        message indicating the successful execution of the job. If the job execution resulted in an exception,
        the error is logged and reported using the application's error handling system. Additionally, if a
        listener is registered for the executed job, this method invokes the listener with the event details.

        Parameters
        ----------
        event : JobExecuted
            An event object containing details about the executed job, including its ID, return value,
            exception (if any), and traceback.

        Returns
        -------
        None
            This method does not return any value. It performs logging, error reporting, and invokes
            any registered listener for the executed job.

        Notes
        -----
        This method is intended for internal use to centralize the handling of job execution events.
        It ensures that job executions are logged, errors are reported, and any custom listeners
        associated with the job are properly notified.
        """

        # Extract the job ID and event code from the event object, using default values if not present
        event_id = self.__getAttribute(event, 'job_id', None)
        event_code = self.__getAttribute(event, 'code', 0)

        # Create an event entity for the executed job, including its ID and code
        data_event = self.__getTaskFromSchedulerById(event_id, event_code)

        # If a listener is registered for this job ID, invoke the listener with the event details
        self.__taskCallableListener(data_event, ListeningEvent.JOB_AFTER)

    def __missedListener(
        self,
        event
    ) -> None:
        """
        Handle job missed events for logging, reporting, and invoking registered listeners.

        This method is triggered when a scheduled job is missed (i.e., it was not executed at its scheduled time)
        by the scheduler. It logs a warning message indicating the missed job and its scheduled run time, creates
        an event entity for the missed job, and invokes any registered listeners for the missed job event. This
        ensures that missed executions are tracked and that any custom logic associated with missed jobs is executed.

        Parameters
        ----------
        event : JobMissed
            An event object containing details about the missed job, including its ID (`job_id`), event code (`code`),
            and the scheduled run time (`scheduled_run_time`).

        Returns
        -------
        None
            This method does not return any value. It performs logging, reporting, and invokes any registered
            listener for the missed job event.

        Notes
        -----
        - This method is intended for internal use to centralize the handling of missed job events.
        - It ensures that missed jobs are logged and that any custom listeners associated with the job are properly notified.
        - The event entity created provides structured information to listeners for further processing.
        """

        # Extract the job ID from the event object, or None if not present
        event_id = self.__getAttribute(event, 'job_id', None)

        # Extract the event code from the event object, defaulting to 0 if not present
        event_code = self.__getAttribute(event, 'code', 0)

        # Create an event entity for the missed job, including its ID and code
        data_event = self.__getTaskFromSchedulerById(event_id, event_code)

        # If a listener is registered for this job ID, invoke the listener with the event details
        self.__taskCallableListener(data_event, ListeningEvent.JOB_ON_MISSED)

    def __maxInstancesListener(
        self,
        event
    ) -> None:
        """
        Handle job max instances events for logging, error reporting, and listener invocation.

        This method is triggered when a job execution exceeds the maximum allowed concurrent instances.
        It logs an error message indicating the job ID that exceeded its instance limit, creates an event
        entity for the affected job, and invokes any registered listener for this job's max instances event.
        This allows for custom handling, notification, or recovery logic to be executed in response to
        the max instances violation.

        Parameters
        ----------
        event : JobMaxInstances
            An event object containing details about the job that exceeded the maximum allowed instances,
            including its job ID and event code.

        Returns
        -------
        None
            This method does not return any value. It performs logging, error reporting, and invokes
            any registered listener for the job max instances event.

        Notes
        -----
        This method is intended for internal use to centralize the handling of job max instances events.
        It ensures that such events are logged and that any custom listeners associated with the job
        are properly notified. No value is returned.
        """

        # Extract the job ID and event code from the event object, using default values if not present
        event_id = self.__getAttribute(event, 'job_id', None)
        event_code = self.__getAttribute(event, 'code', 0)

        # Create an event entity for the job that exceeded max instances
        data_event = self.__getTaskFromSchedulerById(event_id, event_code)

        # If a listener is registered for this job ID, invoke the listener with the event details
        self.__taskCallableListener(data_event, ListeningEvent.JOB_ON_MAXINSTANCES)

    def __removedListener(
        self,
        event
    ) -> None:
        """
        Handle job removal events for logging and invoking registered listeners.

        This method is triggered when a job is removed from the scheduler. It logs an informational
        message indicating that the job has been removed. If the application is in debug mode, it
        may display a message on the console. Additionally, if a listener is registered for the
        removed job, this method invokes the listener with the event details.

        Parameters
        ----------
        event : JobRemoved
            An event object containing details about the removed job, such as its ID and event code.

        Returns
        -------
        None
            This method does not return any value. It performs logging and invokes any registered
            listener for the job removal event.

        Notes
        -----
        - The method retrieves the job ID and event code from the event object.
        - It logs the removal of the job and creates an event entity for the removed job.
        - If a listener is registered for the job, it is invoked with the event details.
        """

        # Retrieve the job ID and event code from the event object
        event_id = self.__getAttribute(event, 'job_id', None)
        event_code = self.__getAttribute(event, 'code', 0)

        # Create an event entity for the removed job
        data_event = self.__getTaskFromSchedulerById(event_id, event_code)

        # If a listener is registered for this job ID, invoke the listener with the event details
        self.__taskCallableListener(data_event, ListeningEvent.JOB_ON_REMOVED)

    def __triggerIsPast(
        self,
        signature: str,
        entity: EventEntity
    ) -> bool:
        """
        Determine if the trigger for a scheduled event is set to a date in the past.

        This method checks whether the trigger associated with the given event entity is a
        `DateTrigger` and, if so, compares its `run_date` to the current time in the scheduler's
        configured timezone. If the trigger's run date is in the past, a warning is logged and
        the method returns True, indicating that the event should be skipped.

        Parameters
        ----------
        signature : str
            The unique signature identifying the scheduled event.
        entity : EventEntity
            The event entity containing the trigger and scheduling information.

        Returns
        -------
        bool
            True if the event's trigger is a `DateTrigger` with a run date in the past; False otherwise.

        Notes
        -----
        This method is used to prevent scheduling events that would immediately be considered expired.
        """

        # Check if the event's trigger is a DateTrigger and its run_date is in the past
        if isinstance(entity.trigger, DateTrigger):
            run_date = getattr(entity.trigger, 'run_date', None)
            now = self.__getNow()

            # If the run_date is set and is before the current time, log a warning and return True
            if run_date is not None and run_date < now:
                self.__logger.warning(
                    f"Scheduled event '{signature}' has a run_date in the past ({run_date}); skipping."
                )
                return True

        # Return False if the trigger is not a past DateTrigger
        return False

    def __eventToEntity(
        self,
        event
    ) -> EventEntity:
        """
        Convert a scheduled event object to its corresponding EventEntity representation.

        This method takes a scheduled event and transforms it into an `EventEntity` instance,
        which encapsulates all relevant attributes required for internal tracking and management
        within the scheduler. The conversion relies on the event implementing a `toEntity` method.

        Parameters
        ----------
        event : Any
            The scheduled event object to be converted. Must implement a `toEntity` method.

        Returns
        -------
        EventEntity
            An instance of `EventEntity` containing the extracted attributes from the event.

        Raises
        ------
        CLIOrionisValueError
            If the provided event does not implement a `toEntity` method.

        Notes
        -----
        This method is used internally to standardize event objects for consistent handling
        within the scheduler's job management system.
        """

        # Ensure the event has a toEntity method for conversion
        if not hasattr(event, 'toEntity'):
            raise CLIOrionisValueError("The event must have a 'toEntity' method for conversion.")

        # Convert the event to its entity representation (EventEntity)
        to_entity = getattr(event, 'toEntity')
        return to_entity()

    def __loadEvents(
        self
    ) -> None:
        """
        Synchronize and register all scheduled events with the AsyncIOScheduler.

        This method updates the internal jobs list (`self.__jobs`) by converting each event in the internal
        events dictionary to its corresponding entity representation. It then schedules these jobs in the
        appropriate scheduler (control or main) based on their signature. Events with triggers set in the
        past are skipped to prevent immediate expiration. Associated listeners are registered for each job
        as needed. This ensures that all scheduled jobs are properly tracked and managed by both the internal
        state and the scheduler.

        Parameters
        ----------
        None

        Returns
        -------
        None
            This method does not return any value. It updates the internal jobs list and registers jobs
            with the AsyncIOScheduler.

        Raises
        ------
        CLIOrionisRuntimeError
            Raised if an error occurs while loading or scheduling an event.

        Notes
        -----
        - Events are loaded only if the internal jobs list is empty to avoid duplicate scheduling.
        - Each event must implement a `toEntity` method for conversion.
        - Jobs are added to the scheduler with their respective configuration, and listeners are registered if present.
        """

        # Only load events if the jobs list is empty to avoid duplicate scheduling
        if not self.__jobs:

            # Lists to separate control jobs and custom jobs
            control_jobs = []
            custom_jobs = []

            # Iterate through all scheduled events in the internal events dictionary
            for signature, event in self.__events.items():

                try:

                    # Convert the event to its entity representation (EventEntity)
                    entity = self.__eventToEntity(event)

                    # Skip loading events with a DateTrigger set in the past
                    if self.__triggerIsPast(signature, entity):
                        continue

                    # Add the job entity to the internal jobs list
                    self.__jobs.append(entity)

                    # Determine which scheduler to use based on the job signature
                    if signature in self.__control_operations:
                        control_jobs.append((signature, entity))
                    else:
                        custom_jobs.append((signature, entity))

                except Exception as e:

                    # Construct an error message for failed event loading
                    error_msg = f"Failed to load scheduled event '{signature}': {str(e)}"

                    # Log the error message
                    self.__logger.error(error_msg)

                    # Raise a runtime error to signal failure in loading the scheduled event
                    raise CLIOrionisRuntimeError(error_msg)

            # Register control jobs with the control scheduler
            self.__loadJobs(self.__control_scheduler, control_jobs)

            # Register custom jobs with the main scheduler
            self.__loadJobs(self.__scheduler, custom_jobs)

            # Log the successful loading of the scheduled events for debugging
            self.__logger.debug(f"Scheduled events loaded successfully: {len(self.__jobs)} jobs.")

    def __loadJobs(
        self,
        scheduler: BaseScheduler,
        events: List[Tuple[str, EventEntity]]
    ) -> None:
        """
        Load and register jobs into the specified scheduler.

        This method iterates over a list of event tuples, each containing a job signature and its corresponding
        EventEntity. For each event, it creates a job function that invokes the reactor with the provided command
        and arguments, then adds the job to the given scheduler with the appropriate configuration. If the event
        entity has an associated listener, the listener is registered for the job.

        Parameters
        ----------
        scheduler : BaseScheduler
            The scheduler instance (either control or main) where jobs will be registered.
        events : List[Tuple[str, EventEntity]]
            A list of tuples, each containing a job signature and its corresponding EventEntity.

        Returns
        -------
        None
            This method does not return any value. It registers jobs and listeners with the scheduler.

        Notes
        -----
        - The job function is created as a lambda that calls the reactor with the command and arguments.
        - Jobs are added with configuration options such as trigger, ID, name, max instances, misfire grace time, and coalesce.
        - If an event entity has a listener, it is registered using the setListener method.
        """

        # Helper function to create a job function that calls the reactor with the given command and arguments
        def create_job_func(cmd, args_list):

            # Returns a lambda that executes the reactor call for the specified command and arguments
            return lambda: self.__reactor.call(cmd, args_list)

        # Iterate over each event tuple and add the job to the scheduler
        for signature, entity in events:

            # Add the job to the scheduler with the specified configuration
            scheduler.add_job(
                func=create_job_func(signature, list(entity.args)),
                trigger=entity.trigger,
                id=signature,
                name=f"{signature}({', '.join(map(str, entity.args))})",
                replace_existing=True,
                max_instances=entity.max_instances,
                misfire_grace_time=entity.misfire_grace_time,
                coalesce=entity.coalesce
            )

            # If the event entity has an associated listener, register it
            if entity.listener:
                self.setListener(signature, entity.listener)

    def __raiseException(
        self,
        exception: BaseException
    ) -> None:
        """
        Centralized exception handling for the scheduler, delegating to the application's error catching system.

        This private method provides a unified mechanism for handling exceptions that occur within the scheduler's
        context. It forwards the exception to the application's error catching mechanism, ensuring consistent
        processing of errors according to the application's global error handling policies. This may include
        logging, reporting, or re-raising the exception, depending on the application's configuration.

        Parameters
        ----------
        exception : BaseException
            The exception instance raised during scheduler or command execution. This can be any subclass of
            BaseException, including system, runtime, or custom exceptions.

        Returns
        -------
        None
            This method always returns None. It delegates the exception to the application's error
            handling system for further processing.

        Notes
        -----
        This method is intended for internal use within the scheduler. The actual handling of the exception
        (such as logging or propagation) is determined by the application's error catching implementation.
        """

        # Create a CLIRequest object representing the current scheduler context
        request = CLIRequest(
            command="schedule:work",
            args={}
        )

        # Delegate the exception to the application's error catching system
        self.__catch.exception(
            self,
            request,
            exception
        )

    def setListener(
        self,
        event: Union[str, ListeningEvent],
        listener: Union[IScheduleEventListener, callable]
    ) -> None:
        """
        Register a listener for a specific scheduler event or job.

        This method allows you to associate a callback or an instance of
        `IScheduleEventListener` with a scheduler event. The event can be a global
        scheduler event (such as 'scheduler_started', 'scheduler_paused', etc.) or
        a specific job ID. When the specified event occurs, the registered listener
        will be invoked with the event data.

        Parameters
        ----------
        event : str or ListeningEvent
            The name of the event to listen for. This can be a string representing
            a global event name (e.g., 'scheduler_started') or a job ID, or an
            instance of `ListeningEvent`.
        listener : IScheduleEventListener or callable
            The listener to be registered. This can be a callable function or an
            instance of `IScheduleEventListener`. The listener should accept the
            event object as its parameter.

        Returns
        -------
        None
            This method does not return any value. It registers the listener for
            the specified event.

        Raises
        ------
        CLIOrionisValueError
            If the event name is not a non-empty string, or if the listener is not
            callable or an instance of `IScheduleEventListener`.

        Notes
        -----
        - If the event parameter is a `ListeningEvent`, its value is used as the event name.
        - The listener will be stored internally and invoked when the corresponding event occurs.
        """

        # If the event is an instance of ListeningEvent, extract its string value
        if isinstance(event, ListeningEvent):
            event = event.value

        # Validate that the event name is a non-empty string
        if not isinstance(event, str) or not event.strip():
            raise CLIOrionisValueError("Event name must be a non-empty string.")

        # Validate that the listener is either callable or an instance of IScheduleEventListener
        if not callable(listener) and not isinstance(listener, IScheduleEventListener):
            raise CLIOrionisValueError(
                "Listener must be a callable function or an instance of IScheduleEventListener."
            )

        # Register the listener for the specified event in the internal listeners dictionary
        self.__listeners[event] = listener

    def pause(
        self
    ) -> None:
        """
        Pause all user jobs managed by the scheduler.

        This method pauses all currently scheduled user jobs in the AsyncIOScheduler if the scheduler is running.
        It iterates through all jobs, attempts to pause each one, and tracks which jobs were paused by adding their
        IDs to an internal set. For each successfully paused job, the method logs the action and invokes any
        registered listeners for the pause event. After all jobs are paused, a global listener for the scheduler
        pause event is also invoked.

        Returns
        -------
        None
            This method does not return any value. It performs the pausing of all user jobs and triggers
            the appropriate listeners and logging.

        Notes
        -----
        - Only jobs with valid IDs are paused.
        - System jobs are not explicitly filtered out; all jobs returned by the scheduler are considered user jobs.
        - If an error occurs while pausing a job, the exception is handled by the application's error handler.
        - The set of paused jobs is cleared before pausing to ensure only currently paused jobs are tracked.
        """

        # Only pause jobs if the scheduler is currently running
        if self.isRunning():

            # Retrieve all jobs currently managed by the scheduler
            all_jobs = self.__scheduler.get_jobs()

            # Iterate through each job and attempt to pause it
            for job in all_jobs:

                try:

                    # Get the job ID safely
                    job_id = self.__getAttribute(job, 'id', None)

                    # Skip jobs without a valid user-defined ID (ignore system/operation jobs)
                    if not job_id or not isinstance(job_id, str) or job_id.strip() == "":
                        continue

                    # Retrieve event data for the paused job
                    event_data = self.__getTaskFromSchedulerById(job_id)

                    # Invoke the listener for the paused job event
                    self.__taskCallableListener(
                        event_data,
                        ListeningEvent.JOB_ON_PAUSED
                    )

                except Exception as e:

                    # Handle any errors that occur while pausing a job
                    self.__raiseException(e)

            # After all jobs are paused, invoke the global listener for scheduler pause
            self.__globalCallableListener(SchedulerPaused(
                code=EVENT_SCHEDULER_PAUSED,
                time=self.__getNow()
            ), ListeningEvent.SCHEDULER_PAUSED)

            # Get the current time in the configured timezone for logging
            now = self.__getCurrentTime()

            # Clear the set of previously paused jobs to avoid stale entries
            self.__scheduler.pause()

            # Store the timestamp when the scheduler was paused
            self.__paused_at = self.__getNow()

            # Log that all tasks have been paused
            self.__logger.info(f"Orionis Scheduler paused all tasks successfully at: {now}.")

    def resume(
        self
    ) -> None:
        """
        Resume all user jobs that were previously paused by the scheduler.

        This method resumes only those jobs that were paused using the `pause` method,
        as tracked by the internal set `__pausedByPauseEverything`. It iterates through
        each paused job, attempts to resume it, and triggers any registered listeners
        for the resumed job event. After all jobs are resumed, a global listener for the
        scheduler resume event is also invoked. The set of paused jobs is cleared after
        resumption to ensure accurate tracking.

        Returns
        -------
        None
            This method does not return any value. It resumes all jobs that were paused
            by the scheduler, triggers the appropriate listeners, and logs the actions.

        Notes
        -----
        - Only jobs that were paused by the `pause` method are resumed.
        - If an error occurs while resuming a job, the exception is handled by the application's error handler.
        - After resuming all jobs, the global scheduler resumed event is triggered.
        """

        # Only resume jobs if the scheduler is currently running
        if self.isPaused():

            # Execute the global callable listener after all jobs are resumed
            self.__globalCallableListener(SchedulerResumed(
                code=EVENT_SCHEDULER_RESUMED,
                time=self.__getNow()
            ), ListeningEvent.SCHEDULER_RESUMED)

            # Get the current time in the configured timezone for logging
            now = self.__getCurrentTime()

            # Cuando resumes:
            resumed_at = self.__getNow()

            # Resume main scheduler
            self.__scheduler.resume()

            # Iterate through all jobs to check for missed executions
            for job in self.__scheduler.get_jobs():

                # If the job was supposed to run while paused, trigger the missed event
                if job.next_run_time and self.__paused_at <= job.next_run_time < resumed_at:

                    # Retrieve event data for the missed job
                    event_data = self.__getTaskFromSchedulerById(job.id)

                    # Invoke the listener for the missed job event
                    self.__taskCallableListener(
                        event_data,
                        ListeningEvent.JOB_ON_MISSED
                    )

            # Log an informational message indicating that the scheduler has been resumed
            self.__logger.info(f"Orionis Scheduler resumed all tasks successfully at: {now}.")

    async def start(
        self
    ) -> None:
        """
        Start the AsyncIO scheduler instance and keep it running.

        This method initializes and starts the AsyncIOScheduler, which integrates with the asyncio event loop
        to manage asynchronous job execution. It ensures that all scheduled events are loaded, listeners are
        subscribed, and the scheduler is started within an asyncio context. The method keeps the scheduler
        running until a stop signal is received, handling graceful shutdowns and interruptions.

        Returns
        -------
        None
            This method does not return any value. It starts the AsyncIO scheduler, keeps it running, and
            ensures proper cleanup during shutdown.

        Raises
        ------
        CLIOrionisRuntimeError
            If the scheduler fails to start due to missing an asyncio event loop or other runtime issues.
        """
        try:

            # Ensure the method is called within an asyncio event loop
            asyncio.get_running_loop()

            # Create an asyncio event to manage clean shutdowns
            self._stop_event = asyncio.Event()

            # Load all scheduled events into the internal jobs list
            self.__loadEvents()

            # Subscribe to scheduler events for monitoring and handling
            self.__subscribeListeners()

            # Start the scheduler if it is not already running
            if not self.isRunning():
                self.__scheduler.start()
                self.__control_scheduler.start()

            # Log that the scheduler is now active and waiting for events
            self.__logger.info("Orionis Scheduler is now active and waiting for events...")

            try:
                # Wait for the stop event to be set, which signals a shutdown
                # This avoids using a busy loop and is more efficient
                await self._stop_event.wait()

            except (KeyboardInterrupt, asyncio.CancelledError):

                # Handle graceful shutdown when an interruption signal is received
                self.__logger.info("Received shutdown signal, stopping scheduler...")
                await self.shutdown(wait=True)

            except Exception as e:

                # Log and raise any unexpected exceptions during scheduler operation
                self.__logger.error(f"Error during scheduler operation: {str(e)}")
                raise CLIOrionisRuntimeError(f"Scheduler operation failed: {str(e)}") from e

            finally:

                # Ensure the scheduler is shut down properly, even if an error occurs
                if self.__scheduler.running:
                    await self.shutdown(wait=False)

        except RuntimeError as e:

            # Handle the case where no asyncio event loop is running
            if "no running event loop" in str(e):
                raise CLIOrionisRuntimeError("Scheduler must be started within an asyncio event loop") from e
            raise CLIOrionisRuntimeError(f"Failed to start the scheduler: {str(e)}") from e

        except Exception as e:

            # Raise a runtime error for any other issues during startup
            raise CLIOrionisRuntimeError(f"Failed to start the scheduler: {str(e)}") from e

    async def shutdown(
        self,
        wait: bool = True
    ) -> None:
        """
        Shut down the AsyncIO scheduler instance asynchronously.

        This method gracefully stops the AsyncIOScheduler and signals the main event loop
        to stop waiting, allowing for clean application shutdown.

        Parameters
        ----------
        wait : bool, optional
            If True, waits for currently executing jobs to complete. Default is True.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the 'wait' parameter is not a boolean value.
        CLIOrionisRuntimeError
            If an error occurs during the shutdown process.
        """

        # Validate the wait parameter
        if not isinstance(wait, bool):
            self.__raiseException(
                CLIOrionisValueError(
                    "The 'wait' parameter must be a boolean value (True or False) to indicate whether to wait for running jobs to finish before shutting down the scheduler."
                )
            )

        # If the scheduler is not running, there's nothing to shut down
        if not self.isRunning():
            return

        try:

            # Shut down the AsyncIOScheduler
            self.__scheduler.shutdown(wait=wait)
            self.__control_scheduler.shutdown(wait=wait)

            # Signal the stop event to break the wait in start()
            if self._stop_event and not self._stop_event.is_set():
                self._stop_event.set()

            # Allow time for cleanup if waiting
            if wait:
                await asyncio.sleep(0.1)

            # Get the current time in the configured timezone for logging
            now = self.__getCurrentTime()

            # Log the successful shutdown
            self.__logger.info(f"Orionis Scheduler has been shut down successfully at: {now}.")

        except Exception as e:

            # Handle exceptions that may occur during shutdown
            self.__raiseException(
                CLIOrionisRuntimeError(
                    f"Error while attempting to shut down Orionis Scheduler: {str(e)}"
                )
            )

    def pauseCommand(
        self,
        signature: str
    ) -> bool:
        """
        Pause a scheduled job in the AsyncIO scheduler.

        This method attempts to pause a job managed by the AsyncIOScheduler, identified by its unique signature.
        It first validates that the provided signature is a non-empty string. If the job exists and is successfully
        paused, the method logs the action and returns True. If the job does not exist or an error occurs during
        the pause operation, the method returns False.

        Parameters
        ----------
        signature : str
            The unique signature (ID) of the job to pause. Must be a non-empty string.

        Returns
        -------
        bool
            True if the job was successfully paused.
            False if the job does not exist or an error occurred during the pause operation.

        Raises
        ------
        CLIOrionisValueError
            If the `signature` parameter is not a non-empty string.

        Notes
        -----
        This method is intended for use with jobs managed by the AsyncIOScheduler. It does not raise
        an exception if the job does not exist; instead, it returns False to indicate failure.
        """

        # Validate that the signature is a non-empty string
        if not isinstance(signature, str) or not signature.strip():
            self.__raiseException(CLIOrionisValueError(self.SIGNATURE_REQUIRED_ERROR))

        try:

            # Attempt to pause the job with the given signature
            self.__scheduler.pause_job(signature)

            # Retrieve event data for the paused job
            event_data = self.__getTaskFromSchedulerById(signature)

            # Invoke the listener for the paused job event
            self.__taskCallableListener(
                event_data,
                ListeningEvent.JOB_ON_PAUSED
            )

            # Log the successful pausing of the job
            self.__logger.info(f"Pause '{signature}' has been paused.")

            # Return True to indicate the job was successfully paused
            return True

        except Exception:

            # Return False if the job could not be paused (e.g., it does not exist or another error occurred)
            return False

    def resumeCommand(
        self,
        signature: str
    ) -> bool:
        """
        Resume a paused job in the AsyncIO scheduler.

        This method attempts to resume a job that was previously paused in the AsyncIOScheduler.
        It first validates that the provided job signature is a non-empty string. If the job exists
        and is currently paused, the method resumes the job, logs the action, and returns True.
        If the job does not exist or an error occurs during the resume operation, the method returns False.

        Parameters
        ----------
        signature : str
            The unique signature (ID) of the job to resume. Must be a non-empty string.

        Returns
        -------
        bool
            True if the job was successfully resumed.
            False if the job does not exist or an error occurred during the resume operation.

        Raises
        ------
        CLIOrionisValueError
            If the `signature` parameter is not a non-empty string.

        Notes
        -----
        This method is intended for use with jobs managed by the AsyncIOScheduler. It does not raise
        an exception if the job does not exist; instead, it returns False to indicate failure.
        """

        # Validate that the signature is a non-empty string
        if not isinstance(signature, str) or not signature.strip():
            self.__raiseException(CLIOrionisValueError(self.SIGNATURE_REQUIRED_ERROR))

        try:

            # Attempt to resume the job with the given signature
            self.__scheduler.resume_job(signature)

            # Retrieve event data for the resumed job
            event_data = self.__getTaskFromSchedulerById(signature)

            # Invoke the listener for the resumed job event
            self.__taskCallableListener(
                event_data,
                ListeningEvent.JOB_ON_RESUMED
            )

            # Log the successful resumption of the job
            self.__logger.info(f"Task '{signature}' has been resumed.")

            # Return True to indicate the job was successfully resumed
            return True

        except Exception:

            # Return False if the job could not be resumed (e.g., it does not exist or another error occurred)
            return False

    def removeCommand(
        self,
        signature: str
    ) -> bool:
        """
        Remove a scheduled job from the AsyncIO scheduler by its signature.

        This method attempts to remove a job from the AsyncIOScheduler using its unique signature (ID).
        It first validates that the provided signature is a non-empty string. If the job exists,
        it is removed from both the scheduler and the internal jobs list. The method logs the removal
        and returns True if successful. If the job does not exist or an error occurs during removal,
        the method returns False.

        Parameters
        ----------
        signature : str
            The unique signature (ID) of the job to remove. Must be a non-empty string.

        Returns
        -------
        bool
            True if the job was successfully removed from both the scheduler and the internal jobs list.
            False if the job does not exist or an error occurred during removal.

        Raises
        ------
        CLIOrionisValueError
            If the `signature` parameter is not a non-empty string.

        Notes
        -----
        This method ensures that both the scheduler and the internal jobs list remain consistent
        after a job is removed. No exception is raised if the job does not exist; instead, False is returned.
        """

        # Validate that the signature is a non-empty string
        if not isinstance(signature, str) or not signature.strip():
            self.__raiseException(CLIOrionisValueError(self.SIGNATURE_REQUIRED_ERROR))

        try:

            # Attempt to remove the job from the scheduler using its signature
            self.__scheduler.remove_job(signature)

            # Remove the job from the internal jobs list if it exists
            for job in self.__jobs:
                if job.signature == signature:
                    self.__jobs.remove(job)
                    break

            # Log the successful removal of the job
            self.__logger.info(f"Task '{signature}' has been removed from the scheduler.")

            # Return True to indicate the job was successfully removed
            return True

        except Exception:

            # Return False if the job could not be removed (e.g., it does not exist or another error occurred)
            return False

    def events(
        self
    ) -> List[Dict]:
        """
        Retrieve a list of all scheduled jobs managed by the scheduler.

        This method ensures that all scheduled events are loaded into the internal jobs list,
        then iterates through each job to collect its details in a dictionary format. Each
        dictionary contains information such as the command signature, arguments, purpose,
        random delay, start and end dates, and any additional job details.

        Returns
        -------
        List[dict]
            A list of dictionaries, where each dictionary represents a scheduled job and contains:
                - 'signature' (str): The command signature of the job.
                - 'args' (list): The arguments passed to the command.
                - 'purpose' (str): The description or purpose of the job.
                - 'random_delay' (int): The random delay associated with the job, if any.
                - 'start_date' (str): The formatted start date and time of the job, or 'Not Applicable' if not set.
                - 'end_date' (str): The formatted end date and time of the job, or 'Not Applicable' if not set.
                - 'details' (str): Additional details about the job.
        Notes
        -----
        This method guarantees that the returned list reflects the current state of all jobs
        managed by this scheduler instance. If no jobs are scheduled, an empty list is returned.
        """

        # Ensure all events are loaded into the internal jobs list
        self.__loadEvents()

        # Initialize a list to hold details of each scheduled job
        events: list = []

        # Iterate over each job in the internal jobs list
        for job in self.__jobs:

            # Safely extract job details with default values if attributes are missing
            signature: str = self.__getAttribute(job, 'signature', '')
            args: list = self.__getAttribute(job, 'args', [])
            purpose: str = self.__getAttribute(job, 'purpose', 'No Description')
            random_delay: int = self.__getAttribute(job, 'random_delay', 0)
            start_date: datetime = self.__getAttribute(job, 'start_date', None)
            end_date: datetime = self.__getAttribute(job, 'end_date', None)
            details: str = self.__getAttribute(job, 'details', 'Not Available')
            coalesce: bool = self.__getAttribute(job, 'coalesce', False)
            max_instances: int = self.__getAttribute(job, 'max_instances', 1)
            misfire_grace_time: int = self.__getAttribute(job, 'misfire_grace_time', None)

            # Format the start and end dates as strings, or mark as 'Not Applicable' if not set
            formatted_start = start_date.strftime('%Y-%m-%d %H:%M:%S') if start_date else self.NOT_APPLICABLE
            formatted_end = end_date.strftime('%Y-%m-%d %H:%M:%S') if end_date else self.NOT_APPLICABLE

            # Append a dictionary with relevant job details to the events list
            events.append({
                'signature': signature,
                'args': args,
                'purpose': purpose,
                'random_delay': random_delay,
                'start_date': formatted_start,
                'end_date': formatted_end,
                'details': details,
                'coalesce': coalesce,
                'max_instances': max_instances,
                'misfire_grace_time': misfire_grace_time
            })

        # Return the list of scheduled job details
        return events

    def event(
        self,
        signature: str
    ) -> Optional[Dict]:
        """
        Retrieve the details of a specific scheduled job by its signature.

        This method searches the internal jobs list for a job whose signature matches
        the provided value. If a matching job is found, it returns a dictionary containing
        the job's details, such as its arguments, purpose, random delay, start and end dates,
        and additional details. If no job with the given signature exists, the method returns None.

        Parameters
        ----------
        signature : str
            The unique signature (ID) of the job to retrieve. Must be a non-empty string.

        Returns
        -------
        dict or None
            If a job with the specified signature is found, returns a dictionary with the following keys:
                - 'signature': str, the job's signature.
                - 'args': list, the arguments passed to the job.
                - 'purpose': str, the description or purpose of the job.
                - 'random_delay': int, the random delay associated with the job (if any).
                - 'start_date': str, the formatted start date and time of the job, or 'Not Applicable' if not set.
                - 'end_date': str, the formatted end date and time of the job, or 'Not Applicable' if not set.
                - 'details': str, additional details about the job.
            If no job with the given signature is found, returns None.

        Notes
        -----
        This method ensures that all events are loaded before searching for the job.
        The returned dictionary provides a summary of the job's configuration and metadata.
        """
        # Ensure all events are loaded into the internal jobs list
        self.__loadEvents()

        # Validate the signature parameter
        if not isinstance(signature, str) or not signature.strip():
            return None

        # Search for the job with the matching signature
        for job in self.__jobs:
            # Get the job's signature attribute safely
            job_signature = self.__getAttribute(job, 'signature', '')

            # If a matching job is found, return its details in a dictionary
            if job_signature == signature:

                # Extract job details safely with default values
                args: list = self.__getAttribute(job, 'args', [])
                purpose: str = self.__getAttribute(job, 'purpose', 'No Description')
                random_delay: int = self.__getAttribute(job, 'random_delay', 0)
                start_date: datetime = self.__getAttribute(job, 'start_date', None)
                end_date: datetime = self.__getAttribute(job, 'end_date', None)
                details: str = self.__getAttribute(job, 'details', 'Not Available')

                # Return the job details as a dictionary
                return {
                    'signature': job_signature,
                    'args': args,
                    'purpose': purpose,
                    'random_delay': random_delay,
                    'start_date': start_date.strftime('%Y-%m-%d %H:%M:%S') if start_date else self.NOT_APPLICABLE,
                    'end_date': end_date.strftime('%Y-%m-%d %H:%M:%S') if end_date else self.NOT_APPLICABLE,
                    'details': details
                }

        # Return None if no job with the given signature is found
        return None

    def isRunning(
        self
    ) -> bool:
        """
        Check if the scheduler is currently running.

        This method inspects the internal state of the AsyncIOScheduler instance to determine
        whether the scheduler is actively running. The scheduler is considered running if it
        has been started and has not been paused or shut down.

        Returns
        -------
        bool
            True if the AsyncIOScheduler is currently running; False otherwise.
        """

        # Return True if the scheduler is running, otherwise False
        return self.__scheduler.running and self.__scheduler.state == STATE_RUNNING

    def isPaused(
        self
    ) -> bool:
        """
        Check if the scheduler is currently paused.

        This method determines whether the scheduler is in a paused state by checking if there are
        any jobs that were paused using the `pause` method. If there are jobs in the internal set
        `__pausedByPauseEverything`, it indicates that the scheduler has been paused.

        Returns
        -------
        bool
            True if the scheduler is currently paused (i.e., there are jobs in the paused set);
            False otherwise.
        """

        # The scheduler is considered paused if there are any jobs in the paused set
        return self.__scheduler.running and self.__scheduler.state == STATE_PAUSED

    def forceStop(
        self
    ) -> None:
        """
        Forcefully stop the scheduler immediately, bypassing graceful shutdown.

        This method immediately shuts down the AsyncIOScheduler instance without waiting for any currently
        running jobs to finish. It is intended for emergency or critical situations where an abrupt stop
        is required, such as unrecoverable errors or forced application termination. In addition to shutting
        down the scheduler, it also signals the internal stop event to interrupt the scheduler's main loop,
        allowing the application to proceed with its shutdown procedures.

        Returns
        -------
        None
            This method does not return any value. It performs a forceful shutdown of the scheduler and
            signals the stop event to ensure the main loop is interrupted.

        Notes
        -----
        - This method should be used with caution, as it does not wait for running jobs to complete.
        - After calling this method, the scheduler will be stopped and any pending or running jobs may be interrupted.
        """

        # If the scheduler is currently running, shut it down immediately without waiting for jobs to finish
        if self.__scheduler.running:
            self.__scheduler.shutdown(wait=False)
            self.__control_scheduler.shutdown(wait=False)

        # If the stop event exists and has not already been set, signal it to interrupt the main loop
        if self._stop_event and not self._stop_event.is_set():
            self._stop_event.set()

    def stop(
        self
    ) -> None:
        """
        Signal the scheduler to stop synchronously by setting the internal stop event.

        This method is used to request a graceful shutdown of the scheduler from a synchronous context.
        It sets the internal asyncio stop event, which will cause the scheduler's main loop to exit.
        If an asyncio event loop is currently running, the stop event is set in a thread-safe manner
        using `call_soon_threadsafe`. If no event loop is running, the stop event is set directly.
        This method is safe to call from both asynchronous and synchronous contexts.

        Returns
        -------
        None
            This method does not return any value. It only signals the scheduler to stop by setting
            the internal stop event.

        Notes
        -----
        - If the stop event is already set or does not exist, this method does nothing.
        - Any exceptions encountered while setting the stop event are logged as warnings, but the
          method will still attempt to set the event directly.
        """
        # Check if the stop event exists and has not already been set
        if self._stop_event and not self._stop_event.is_set():

            try:

                # Try to get the current running event loop
                loop = asyncio.get_running_loop()

                # If the event loop is running, set the stop event in a thread-safe manner
                loop.call_soon_threadsafe(self._stop_event.set)

            except RuntimeError:

                # No running event loop, set the stop event directly
                self._stop_event.set()

            except Exception as e:

                # Log any unexpected error but still try to set the event directly
                self.__logger.warning(f"Error setting stop event through event loop: {str(e)}")
                self._stop_event.set()


from orionis.test.exceptions import OrionisTestValueError

class __ValidModuleName:

    def __call__(self, module_name) -> str:
        """
        Validates that the input is a non-empty string.

        Parameters
        ----------
        module_name : Any
            The value to validate as a non-empty string.

        Returns
        -------
        str
            The validated and stripped string value of `module_name`.

        Raises
        ------
        OrionisTestValueError
            If `module_name` is not a non-empty string.
        """
        if not isinstance(module_name, str) or not module_name.strip():
            raise OrionisTestValueError(
                f"Invalid module_name: Expected a non-empty string, got '{module_name}' ({type(module_name).__name__})."
            )
        return module_name

# Exported singleton instance
ValidModuleName = __ValidModuleName()


from typing import Any, Optional
from orionis.services.environment.contracts.caster import IEnvironmentCaster
from orionis.services.environment.enums.value_type import EnvironmentValueType
from orionis.services.environment.exceptions import OrionisEnvironmentValueError, OrionisEnvironmentValueException

class EnvironmentCaster(IEnvironmentCaster):

    # Type class to handle different types of environment variables
    OPTIONS = {e.value for e in EnvironmentValueType}

    @staticmethod
    def options() -> set: # NOSONAR
        """
        Get the set of valid type hints supported by this class.

        Returns
        -------
        set
            Set of valid type hint strings that can be used for environment value casting.
        """
        return EnvironmentCaster.OPTIONS

    def __init__(
        self,
        raw: str | Any
    ) -> None:
        """
        Parse the input `raw` to extract a type hint and value for environment variable casting.

        Parameters
        ----------
        raw : str or Any
            The input value to be parsed. If a string, it may contain a type hint and value
            separated by a colon (e.g., "int: 42"). If a colon is present, the part before
            the colon is treated as the type hint and the part after as the value. If no colon
            is present, the entire string is treated as the value with no type hint. If not a
            string, the input is treated as the value with no type hint.

        Attributes
        ----------
        __type_hint : str or None
            The extracted type hint in lowercase, or None if not provided or invalid.
        __value_raw : str or Any
            The extracted value string if input is a string, or the raw value otherwise.

        Notes
        -----
        This constructor does not return a value. It initializes the instance attributes
        for type hint and raw value, which are used for subsequent type casting operations.
        """

        # Initialize type hint and value to default None
        self.__type_hint: Optional[str] = None
        self.__value_raw: str | Any = None

        # If the input is a string, attempt to parse type hint and value
        if isinstance(raw, str):

            # Remove leading whitespace from the input
            self.__value_raw = raw.lstrip()

            # Check if the string contains a colon, indicating a type hint
            if ':' in self.__value_raw:

                # Split at the first colon to separate type hint and value
                type_hint, value_str = raw.split(':', 1)

                # Validate the extracted type hint and set attributes if valid
                if type_hint.strip().lower() in self.OPTIONS:
                    self.__type_hint = type_hint.strip().lower()
                    # Remove leading whitespace from the value part
                    self.__value_raw = value_str.lstrip() if value_str else None
        else:

            # If input is not a string, treat it as the value with no type hint
            self.__value_raw = raw

    def get(self): # NOSONAR
        """
        Returns the processed value based on the specified type hint.

        If a valid type hint is present, this method dispatches to the corresponding
        internal parsing method for that type. Supported type hints include: 'path',
        'str', 'int', 'float', 'bool', 'list', 'dict', 'tuple', 'set', and 'base64'.
        If no type hint is set, the raw value is returned as is.

        Returns
        -------
        Any
            The value converted or processed according to the specified type hint.
            If no type hint is set, returns the raw value.

        Raises
        ------
        OrionisEnvironmentValueError
            If an error occurs during type conversion or processing.
        """

        # Attempt to process the value based on the type hint
        try:

            # If a type hint is set, dispatch to the appropriate parsing method
            if self.__type_hint:

                # Handle 'path' type hint
                if self.__type_hint == EnvironmentValueType.PATH.value:
                    return self.__parsePath()

                # Handle 'str' type hint
                if self.__type_hint == EnvironmentValueType.STR.value:
                    return self.__parseStr()

                # Handle 'int' type hint
                if self.__type_hint == EnvironmentValueType.INT.value:
                    return self.__parseInt()

                # Handle 'float' type hint
                if self.__type_hint == EnvironmentValueType.FLOAT.value:
                    return self.__parseFloat()

                # Handle 'bool' type hint
                if self.__type_hint == EnvironmentValueType.BOOL.value:
                    return self.__parseBool()

                # Handle 'list' type hint
                if self.__type_hint == EnvironmentValueType.LIST.value:
                    return self.__parseList()

                # Handle 'dict' type hint
                if self.__type_hint == EnvironmentValueType.DICT.value:
                    return self.__parseDict()

                # Handle 'tuple' type hint
                if self.__type_hint == EnvironmentValueType.TUPLE.value:
                    return self.__parseTuple()

                # Handle 'set' type hint
                if self.__type_hint == EnvironmentValueType.SET.value:
                    return self.__parseSet()

                # Handle 'base64' type hint
                if self.__type_hint == EnvironmentValueType.BASE64.value:
                    return self.__parseBase64()

            else:

                # If no type hint is set, return the raw value
                return self.__value_raw

        except OrionisEnvironmentValueError:

            # Propagate specific type conversion errors
            raise

        except Exception as e:

            # Catch any other unexpected errors and wrap them in an environment value error
            raise OrionisEnvironmentValueError(
                f"Error processing value '{self.__value_raw}' with type hint '{self.__type_hint}': {str(e)}"
            ) from e

    def to(self, type_hint: str | EnvironmentValueType) -> Any:
        """
        Convert the internal value to the specified type and return its string representation
        with the type hint prefix.

        Parameters
        ----------
        type_hint : str or EnvironmentValueType
            The type to which the internal value should be converted. Can be a string or an
            EnvironmentValueType enum member. Must be one of the valid options in `OPTIONS`.

        Returns
        -------
        Any
            The string representation of the value with the type hint prefix, according to the
            specified type. For example, "int:42", "list:[1, 2, 3]", etc.

        Raises
        ------
        OrionisEnvironmentValueError
            If the provided type hint is not valid or if the value cannot be converted to the
            specified type.
        """

        # Validate the type hint and ensure it is one of the defined options
        try:

            # If type_hint is an enum, convert it to its value string
            if isinstance(type_hint, EnvironmentValueType):
                type_hint = type_hint.value

            # Validate the type hint against the defined options
            if type_hint not in self.OPTIONS:
                raise OrionisEnvironmentValueError(
                    f"Invalid type hint: {type_hint}. Must be one of {self.OPTIONS}."
                )

            # Set the type hint for the instance
            self.__type_hint = type_hint

            # Dispatch to the appropriate conversion method based on the type hint
            if self.__type_hint == EnvironmentValueType.PATH.value:
                return self.__toPath()

            # If the type hint is 'str', convert to string
            if self.__type_hint == EnvironmentValueType.STR.value:
                return self.__toStr()

            # If the type hint is 'int', convert to integer
            if self.__type_hint == EnvironmentValueType.INT.value:
                return self.__toInt()

            # If the type hint is 'float', convert to float
            if self.__type_hint == EnvironmentValueType.FLOAT.value:
                return self.__toFloat()

            # If the type hint is 'bool', convert to boolean
            if self.__type_hint == EnvironmentValueType.BOOL.value:
                return self.__toBool()

            # If the type hint is 'list', convert to list
            if self.__type_hint == EnvironmentValueType.LIST.value:
                return self.__toList()

            # If the type hint is 'dict', convert to dictionary
            if self.__type_hint == EnvironmentValueType.DICT.value:
                return self.__toDict()

            # If the type hint is 'tuple', convert to tuple
            if self.__type_hint == EnvironmentValueType.TUPLE.value:
                return self.__toTuple()

            # If the type hint is 'set', convert to set
            if self.__type_hint == EnvironmentValueType.SET.value:
                return self.__toSet()

            # If the type hint is 'base64', convert to Base64 encoded string
            if self.__type_hint == EnvironmentValueType.BASE64.value:
                return self.__toBase64()

        except OrionisEnvironmentValueError:

            # Propagate specific type conversion errors
            raise

        except Exception as e:

            # Catch any other unexpected errors and wrap them in an environment value error
            raise OrionisEnvironmentValueError(
                f"Error converting value '{self.__value_raw}' to type '{type_hint}': {str(e)}"
            ) from e

    def __toBase64(self) -> str:
        """
        Convert the internal value to a Base64 encoded string with the type hint prefix.

        - If the value is already valid base64, leave it as-is.
        - Otherwise, encode it to base64.

        Returns
        -------
        str
            A string in the format "<type_hint>:<base64_value>".
        """
        import base64

        if not isinstance(self.__value_raw, (str, bytes)):
            raise OrionisEnvironmentValueError(
                f"Value must be a string or bytes to convert to Base64, got {type(self.__value_raw).__name__} instead."
            )

        # Normalizar a str
        if isinstance(self.__value_raw, bytes):
            candidate = self.__value_raw.decode("utf-8", errors="ignore")
        else:
            candidate = self.__value_raw

        try:
            # Validar si ya es base64 válido
            base64.b64decode(candidate, validate=True)
            encoded_value = candidate
        except Exception:
            # No era base64, entonces convertirlo
            raw_bytes = (
                self.__value_raw.encode() if isinstance(self.__value_raw, str) else self.__value_raw
            )
            encoded_value = base64.b64encode(raw_bytes).decode("utf-8")

        return f"{self.__type_hint}:{encoded_value}"

    def __parseBase64(self) -> bytes:
        """
        Decode the internal raw value from Base64 encoding.

        Returns
        -------
        bytes
            The decoded raw bytes from the Base64-encoded internal value.

        Raises
        ------
        OrionisEnvironmentValueException
            If the internal value cannot be decoded from Base64.
        """
        import base64

        try:
            raw_value = self.__value_raw
            if isinstance(raw_value, bytes):
                raw_value = raw_value.decode("utf-8", errors="ignore")

            decoded = base64.b64decode(raw_value, validate=True)

            try:
                return decoded.decode("utf-8")
            except UnicodeDecodeError:
                return decoded

        except Exception as e:
            raise OrionisEnvironmentValueException(
                f"Cannot decode Base64 value '{self.__value_raw}': {str(e)}"
            )

    def __parsePath(self):
        """
        Convert the internal raw value to a normalized POSIX path string.

        This method processes the internal value as a file system path. If the value is already
        a `Path` object, it returns its POSIX representation. If the value is a string, it replaces
        backslashes with forward slashes for normalization and returns the POSIX path string.

        Returns
        -------
        str
            The normalized POSIX path string representing the file system path.

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be processed as a valid path.
        """

        # Import the Path class from pathlib for path manipulation
        from pathlib import Path

        # If the value is already a Path object, return its POSIX representation
        if isinstance(self.__value_raw, Path):
            return self.__value_raw.as_posix()

        # Normalize the path by replacing backslashes with forward slashes
        normalized_path = str(self.__value_raw).replace('\\', '/')

        # Convert the normalized string to a Path object and return its POSIX representation
        return Path(normalized_path).as_posix()

    def __toPath(self) -> str:
        """
        Convert the internal value to an absolute POSIX path string with type hint prefix.

        Returns
        -------
        str
            A string in the format "<type_hint>:<absolute_path>", where <absolute_path> is the
            normalized, absolute POSIX path representation of the internal value.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value is not a string or a pathlib.Path object.
        """

        # Import the Path class from pathlib for path manipulation
        from pathlib import Path

        # Ensure the internal value is a string or Path object
        if not isinstance(self.__value_raw, (str, Path)):
            raise OrionisEnvironmentValueError(
                f"Value must be a string or Path to convert to path, got {type(self.__value_raw).__name__} instead."
            )

        # Normalize slashes and strip whitespace from the path string
        raw_path = str(self.__value_raw).replace('\\', '/').strip()

        # Create a Path object from the normalized path string
        path_obj = Path(raw_path)

        # If the path is not absolute, resolve it relative to the current working directory
        if not path_obj.is_absolute():

            # Remove any leading slash to avoid creating an absolute path when joining
            raw_path_no_leading = raw_path.lstrip('/\\')

            # Combine with the current working directory to form an absolute path
            path_obj = Path(Path.cwd()) / raw_path_no_leading

        # Expand user home and convert to POSIX format for consistency
        abs_path = path_obj.expanduser().as_posix()

        # Return the absolute path as a string with the type hint prefix
        return f"{self.__type_hint}:{str(abs_path)}"

    def __parseStr(self):
        """
        Returns the internal raw value as a string, removing leading whitespace.

        Parameters
        ----------
        self : EnvironmentCaster
            Instance of the EnvironmentCaster class.

        Returns
        -------
        str
            The internal value as a string with leading whitespace removed.

        Notes
        -----
        No type conversion is performed; the value is returned as a string after
        stripping leading whitespace. This method assumes the type hint is 'str:'.
        No exceptions are raised by this method.
        """

        # Remove leading whitespace from the internal value
        # This ensures that any accidental spaces before the value are ignored
        return self.__value_raw.lstrip()

    def __toStr(self):
        """
        Returns the internal value as a string representation with the type hint prefix.

        Returns
        -------
        str
            The string representation of the internal value, prefixed by the type hint and separated by a colon.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value is not a string.
        """

        # Ensure the internal value is a string before conversion
        if not isinstance(self.__value_raw, str):

            # Raise an error if the value is not a string
            raise OrionisEnvironmentValueError(
                f"Value must be a string to convert to str, got {type(self.__value_raw).__name__} instead."
            )

        # Return the formatted string with type hint and value
        return f"{self.__type_hint}:{self.__value_raw}"

    def __parseInt(self):
        """
        Convert the internal raw value to an integer.

        Strips leading and trailing whitespace from the internal raw value and attempts to convert it to an integer.
        Raises an OrionisEnvironmentValueException if the conversion fails due to invalid format or type.

        Returns
        -------
        int
            The internal value converted to an integer.

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be converted to an integer due to invalid format or type.
        """

        # Remove leading and trailing whitespace from the raw value
        value = self.__value_raw.strip()

        # Attempt to convert the value to an integer
        try:
            return int(value)
        # Raise a custom exception if conversion fails
        except ValueError as e:
            raise OrionisEnvironmentValueException(f"Cannot convert '{value}' to int: {str(e)}")

    def __toInt(self):
        """
        Converts the internal value to a string representation with the integer type hint prefix.

        Now supports conversion from string values to integers for better usability.

        Returns
        -------
        str
            String in the format "<type_hint>:<value>", where <type_hint> is the current type hint
            and <value> is the integer value.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value cannot be converted to an integer.
        """

        # If the internal value is already an integer, use it directly
        if isinstance(self.__value_raw, int):
            return f"{self.__type_hint}:{str(self.__value_raw)}"

        # If the internal value is a string, try to convert it to an integer
        if isinstance(self.__value_raw, str):
            try:
                # Strip whitespace and attempt conversion
                converted_value = int(self.__value_raw.strip())
                return f"{self.__type_hint}:{str(converted_value)}"
            except ValueError:
                raise OrionisEnvironmentValueError(
                    f"Cannot convert string '{self.__value_raw}' to integer. Value must be a valid integer representation."
                )

        # For other types, try direct conversion
        try:
            converted_value = int(self.__value_raw)
            return f"{self.__type_hint}:{str(converted_value)}"
        except (ValueError, TypeError):
            raise OrionisEnvironmentValueError(
                f"Value must be convertible to integer, got {type(self.__value_raw).__name__} with value '{self.__value_raw}'."
            )

    def __parseFloat(self):
        """
        Convert the internal raw value to a float.

        Strips leading and trailing whitespace from the internal raw value and attempts to convert it to a float.
        Raises an OrionisEnvironmentValueException if the conversion fails due to invalid format or type.

        Returns
        -------
        float
            The internal value converted to a float.

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be converted to a float due to invalid format or type.
        """

        # Remove leading and trailing whitespace from the raw value to ensure clean input
        value = self.__value_raw.strip()

        # Attempt to convert the value to a float
        try:
            return float(value)

        # Raise a custom exception if conversion fails
        except ValueError as e:
            raise OrionisEnvironmentValueException(f"Cannot convert '{value}' to float: {str(e)}")

    def __toFloat(self):
        """
        Converts the internal value to a string representation with the float type hint prefix.

        Now supports conversion from string values to floats for better usability.

        Returns
        -------
        str
            A string in the format "<type_hint>:<value>", where <type_hint> is the current type hint
            and <value> is the float value.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value cannot be converted to a float.
        """

        # If the internal value is already a float, use it directly
        if isinstance(self.__value_raw, float):
            return f"{self.__type_hint}:{str(self.__value_raw)}"

        # If the internal value is a string, try to convert it to a float
        if isinstance(self.__value_raw, str):
            try:
                # Strip whitespace and attempt conversion
                converted_value = float(self.__value_raw.strip())
                return f"{self.__type_hint}:{str(converted_value)}"
            except ValueError:
                raise OrionisEnvironmentValueError(
                    f"Cannot convert string '{self.__value_raw}' to float. Value must be a valid floating-point representation."
                )

        # For other types (like int), try direct conversion
        try:
            converted_value = float(self.__value_raw)
            return f"{self.__type_hint}:{str(converted_value)}"
        except (ValueError, TypeError):
            raise OrionisEnvironmentValueError(
                f"Value must be convertible to float, got {type(self.__value_raw).__name__} with value '{self.__value_raw}'."
            )

    def __parseBool(self):
        """
        Convert the internal raw value to a boolean.

        This method strips leading and trailing whitespace from the internal raw value,
        converts it to lowercase, and checks if it matches 'true' or 'false'. If the value
        is 'true', it returns True. If the value is 'false', it returns False. If the value
        does not match either, an OrionisEnvironmentValueException is raised.

        Returns
        -------
        bool
            True if the value is 'true' (case-insensitive), False if the value is 'false' (case-insensitive).

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be converted to a boolean because it does not match 'true' or 'false'.
        """

        # Remove leading and trailing whitespace, then convert to lowercase for comparison
        value = self.__value_raw.strip().lower()

        # If the value is 'true', return True
        if value == 'true':
            return True

        # If the value is 'false', return False
        elif value == 'false':
            return False

        # If the value is neither 'true' nor 'false', raise an exception
        else:
            raise OrionisEnvironmentValueException(f"Cannot convert '{value}' to bool.")

    def __toBool(self):
        """
        Convert the internal value to a string representation with the boolean type hint prefix.

        Now supports conversion from string values to booleans for better usability.

        Returns
        -------
        str
            A string in the format "<type_hint>:<value>", where <type_hint> is the current type hint
            and <value> is the lowercase string representation of the boolean value ('true' or 'false').

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value cannot be converted to a boolean.
        """

        # If the internal value is already a boolean, use it directly
        if isinstance(self.__value_raw, bool):
            return f"{self.__type_hint}:{str(self.__value_raw).lower()}"

        # If the internal value is a string, try to convert it to a boolean
        if isinstance(self.__value_raw, str):
            # Strip whitespace and check common boolean representations
            str_value = self.__value_raw.strip().lower()

            if str_value in ('true', '1', 'yes', 'on', 'enabled'):
                return f"{self.__type_hint}:true"
            elif str_value in ('false', '0', 'no', 'off', 'disabled'):
                return f"{self.__type_hint}:false"
            else:
                raise OrionisEnvironmentValueError(
                    f"Cannot convert string '{self.__value_raw}' to boolean. "
                    f"Valid representations: true/false, 1/0, yes/no, on/off, enabled/disabled."
                )

        # For other types, try direct conversion using Python's truthiness
        try:
            boolean_value = bool(self.__value_raw)
            return f"{self.__type_hint}:{str(boolean_value).lower()}"
        except Exception:
            raise OrionisEnvironmentValueError(
                f"Value must be convertible to boolean, got {type(self.__value_raw).__name__} with value '{self.__value_raw}'."
            )

    def __parseList(self):
        """
        Parses the internal raw value and converts it to a Python list.

        The method strips leading and trailing whitespace from the internal raw value,
        then attempts to safely evaluate the string as a Python list using `ast.literal_eval`.
        If the conversion is successful and the result is a list, it is returned.
        If the conversion fails or the evaluated value is not a list, an
        `OrionisEnvironmentValueException` is raised.

        Returns
        -------
        list
            The internal value converted to a list.

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be converted to a list due to invalid format or type.
        """

        # Import the ast module for safe evaluation of string literals
        import ast

        # Remove leading and trailing whitespace from the raw value to ensure clean input
        value = self.__value_raw.strip()

        try:

            # Safely evaluate the string to a Python object using ast.literal_eval
            parsed = ast.literal_eval(value)

            # Ensure the evaluated object is a list
            if not isinstance(parsed, list):
                raise OrionisEnvironmentValueError("Value is not a list")

            # Return the parsed list if successful
            return parsed

        except (OrionisEnvironmentValueError, ValueError, SyntaxError) as e:

            # Raise a custom exception if conversion fails
            raise OrionisEnvironmentValueException(f"Cannot convert '{value}' to list: {str(e)}")

    def __toList(self):
        """
        Converts the internal value to its string representation with the list type hint prefix.

        Parameters
        ----------
        self : EnvironmentCaster
            Instance of the EnvironmentCaster class.

        Returns
        -------
        str
            A string in the format "<type_hint>:<value>", where <type_hint> is the current type hint
            and <value> is the string representation of the list.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value is not a list.
        """

        # Ensure the internal value is a list before conversion
        if not isinstance(self.__value_raw, list):

            # Raise an error if the value is not a list
            raise OrionisEnvironmentValueError(
                f"Value must be a list to convert to list, got {type(self.__value_raw).__name__} instead."
            )

        # Return the formatted string with type hint and list value
        return f"{self.__type_hint}:{repr(self.__value_raw)}"

    def __parseDict(self):
        """
        Parses the internal raw value and converts it to a Python dictionary.

        The method strips leading and trailing whitespace from the internal raw value,
        then attempts to safely evaluate the string as a Python dictionary using
        `ast.literal_eval`. If the conversion is successful and the result is a dictionary,
        it is returned. If the conversion fails or the evaluated value is not a dictionary,
        an OrionisEnvironmentValueException is raised.

        Returns
        -------
        dict
            The internal value converted to a dictionary.

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be converted to a dictionary due to invalid format or type.
        """

        # Import the ast module for safe evaluation of string literals
        import ast

        # Remove leading and trailing whitespace from the raw value to ensure clean input
        value = self.__value_raw.strip()

        # Attempt to parse the string as a dictionary
        try:

            # Safely evaluate the string to a Python object using ast.literal_eval
            parsed = ast.literal_eval(value)

            # Ensure the evaluated object is a dictionary
            if not isinstance(parsed, dict):
                raise OrionisEnvironmentValueError("Value is not a dict")

            # Return the parsed dictionary if successful
            return parsed

        except (OrionisEnvironmentValueError, ValueError, SyntaxError) as e:

            # Raise a custom exception if conversion fails
            raise OrionisEnvironmentValueException(f"Cannot convert '{value}' to dict: {str(e)}")

    def __toDict(self):
        """
        Converts the internal value to a string representation with the dictionary type hint prefix.

        Returns
        -------
        str
            A string in the format "<type_hint>:<value>", where <type_hint> is the current type hint
            and <value> is the string representation of the dictionary.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value is not a dictionary.
        """

        # Ensure the internal value is a dictionary before conversion
        if not isinstance(self.__value_raw, dict):

            # Raise an error if the value is not a dictionary
            raise OrionisEnvironmentValueError(
                f"Value must be a dict to convert to dict, got {type(self.__value_raw).__name__} instead."
            )

        # Return the formatted string with type hint and dictionary value
        return f"{self.__type_hint}:{repr(self.__value_raw)}"

    def __parseTuple(self):
        """
        Parse the internal raw value and convert it to a Python tuple.

        The method removes leading and trailing whitespace from the internal raw value,
        then attempts to safely evaluate the string as a Python tuple using `ast.literal_eval`.
        If the conversion is successful and the result is a tuple, it is returned.
        If the conversion fails or the evaluated value is not a tuple, an
        OrionisEnvironmentValueException is raised.

        Returns
        -------
        tuple
            The internal value converted to a tuple.

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be converted to a tuple due to invalid format or type.
        """

        # Import the ast module for safe evaluation of string literals
        import ast

        # Remove leading and trailing whitespace from the raw value to ensure clean input
        value = self.__value_raw.strip()

        try:

            # Safely evaluate the string to a Python object using ast.literal_eval
            parsed = ast.literal_eval(value)

            # Ensure the evaluated object is a tuple
            if not isinstance(parsed, tuple):
                raise OrionisEnvironmentValueError("Value is not a tuple")

            # Return the parsed tuple if successful
            return parsed

        except (OrionisEnvironmentValueError, ValueError, SyntaxError) as e:

            # Raise a custom exception if conversion fails
            raise OrionisEnvironmentValueException(f"Cannot convert '{value}' to tuple: {str(e)}")

    def __toTuple(self):
        """
        Convert the internal value to a string representation with the tuple type hint prefix.

        Returns
        -------
        str
            A string in the format "<type_hint>:<value>", where <type_hint> is the current type hint
            and <value> is the string representation of the tuple.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value is not a tuple.
        """

        # Ensure the internal value is a tuple before conversion
        if not isinstance(self.__value_raw, tuple):

            # Raise an error if the value is not a tuple
            raise OrionisEnvironmentValueError(
                f"Value must be a tuple to convert to tuple, got {type(self.__value_raw).__name__} instead."
            )

        # Return the formatted string with type hint and tuple value
        return f"{self.__type_hint}:{repr(self.__value_raw)}"

    def __parseSet(self):
        """
        Parse the internal raw value and convert it to a Python set.

        This method removes leading and trailing whitespace from the internal raw value,
        then attempts to safely evaluate the string as a Python set using `ast.literal_eval`.
        If the conversion is successful and the result is a set, it is returned.
        If the conversion fails or the evaluated value is not a set, an
        OrionisEnvironmentValueException is raised.

        Returns
        -------
        set
            The internal value converted to a set.

        Raises
        ------
        OrionisEnvironmentValueException
            If the value cannot be converted to a set due to invalid format or type.
        """

        # Import the ast module for safe evaluation of string literals
        import ast

        # Remove leading and trailing whitespace from the raw value to ensure clean input
        value = self.__value_raw.strip()

        try:

            # Safely evaluate the string to a Python object using ast.literal_eval
            parsed = ast.literal_eval(value)

            # Ensure the evaluated object is a set
            if not isinstance(parsed, set):
                raise OrionisEnvironmentValueError("Value is not a set")

            # Return the parsed set if successful
            return parsed

        except (OrionisEnvironmentValueError, ValueError, SyntaxError) as e:

            # Raise a custom exception if conversion fails
            raise OrionisEnvironmentValueException(f"Cannot convert '{value}' to set: {str(e)}")

    def __toSet(self):
        """
        Convert the internal value to a string representation with the set type hint prefix.

        Returns
        -------
        str
            A string in the format "<type_hint>:<value>", where <type_hint> is the current type hint
            and <value> is the string representation of the set.

        Raises
        ------
        OrionisEnvironmentValueError
            If the internal value is not a set.
        """

        # Ensure the internal value is a set before conversion
        if not isinstance(self.__value_raw, set):
            # Raise an error if the value is not a set
            raise OrionisEnvironmentValueError(
            f"Value must be a set to convert to set, got {type(self.__value_raw).__name__} instead."
            )

        # Return the formatted string with type hint and set value
        return f"{self.__type_hint}:{repr(self.__value_raw)}"
